from ..utils._scatterplot import _embedding
import collections.abc as cabc
from copy import copy
from numbers import Integral
from itertools import combinations, product
import matplotlib
from typing import (
    Collection,
    Union,
    Optional,
    Sequence,
    Any,
    Mapping,
    List,
    Tuple,
    Literal,
)
from warnings import warn

import matplotlib.patches as mpatches
from scipy.stats import kruskal
import numpy as np
import pandas as pd
from anndata import AnnData
from cycler import Cycler
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from pandas.api.types import is_categorical_dtype
from matplotlib import pyplot as pl, colors, colormaps
from matplotlib import rcParams
from matplotlib import patheffects
from matplotlib.colors import Colormap, Normalize
from functools import partial
import matplotlib.pyplot as plt
import seaborn as sns
import scanpy as sc

from scanpy.plotting import _utils 

from scanpy.plotting._utils import (
    _FontWeight,
    _FontSize,
    ColorLike,
    VBound,
)

def embedding(
    adata: AnnData,
    basis: str,
    *,
    color: Union[str, Sequence[str], None] = None,
    gene_symbols: Optional[str] = None,
    use_raw: Optional[bool] = None,
    sort_order: bool = True,
    edges: bool = False,
    edges_width: float = 0.1,
    edges_color: Union[str, Sequence[float], Sequence[str]] = 'grey',
    neighbors_key: Optional[str] = None,
    arrows: bool = False,
    arrows_kwds: Optional[Mapping[str, Any]] = None,
    groups: Optional[str] = None,
    components: Union[str, Sequence[str]] = None,
    dimensions: Optional[Union[Tuple[int, int], Sequence[Tuple[int, int]]]] = None,
    layer: Optional[str] = None,
    projection: Literal['2d', '3d'] = '2d',
    scale_factor: Optional[float] = None,
    color_map: Union[Colormap, str, None] = None,
    cmap: Union[Colormap, str, None] = None,
    palette: Union[str, Sequence[str], Cycler, None] = None,
    na_color: ColorLike = "lightgray",
    na_in_legend: bool = True,
    size: Union[float, Sequence[float], None] = None,
    frameon: Optional[bool] = 'small',
    legend_fontsize: Union[int, float, _FontSize, None] = None,
    legend_fontweight: Union[int, _FontWeight] = 'bold',
    legend_loc: str = 'right margin',
    legend_fontoutline: Optional[int] = None,
    colorbar_loc: Optional[str] = "right",
    vmax: Union[VBound, Sequence[VBound], None] = None,
    vmin: Union[VBound, Sequence[VBound], None] = None,
    vcenter: Union[VBound, Sequence[VBound], None] = None,
    norm: Union[Normalize, Sequence[Normalize], None] = None,
    add_outline: Optional[bool] = False,
    outline_width: Tuple[float, float] = (0.3, 0.05),
    outline_color: Tuple[str, str] = ('black', 'white'),
    ncols: int = 4,
    hspace: float = 0.25,
    wspace: Optional[float] = None,
    title: Union[str, Sequence[str], None] = None,
    show: Optional[bool] = None,
    save: Union[bool, str, None] = None,
    ax: Optional[Axes] = None,
    return_fig: Optional[bool] = None,
    marker: Union[str, Sequence[str]] = '.',
    **kwargs,
) -> Union[Figure, Axes, None]:
    """\
    Scatter plot for user specified embedding basis (e.g. umap, pca, etc)

    Arguments:
        adata: Annotated data matrix.
        basis: Name of the `obsm` basis to use.
        
    Returns:
        If `show==False` a :class:`~matplotlib.axes.Axes` or a list of it.
    """

    return _embedding(adata=adata, basis=basis, color=color, 
                     gene_symbols=gene_symbols, use_raw=use_raw, 
                     sort_order=sort_order, edges=edges, 
                     edges_width=edges_width, edges_color=edges_color, 
                     neighbors_key=neighbors_key, arrows=arrows, 
                     arrows_kwds=arrows_kwds, groups=groups, 
                     components=components, dimensions=dimensions, 
                     layer=layer, projection=projection, scale_factor=scale_factor,
                       color_map=color_map, cmap=cmap, palette=palette, 
                       na_color=na_color, na_in_legend=na_in_legend, 
                       size=size, frameon=frameon, legend_fontsize=legend_fontsize, 
                       legend_fontweight=legend_fontweight, legend_loc=legend_loc, 
                       legend_fontoutline=legend_fontoutline, colorbar_loc=colorbar_loc, 
                       vmax=vmax, vmin=vmin, vcenter=vcenter, norm=norm, 
                       add_outline=add_outline, outline_width=outline_width, 
                       outline_color=outline_color, ncols=ncols, hspace=hspace,
                         wspace=wspace, title=title, show=show, save=save, ax=ax,
                           return_fig=return_fig, marker=marker, **kwargs)



def cellproportion(adata:AnnData,celltype_clusters:str,groupby:str,
                       groupby_li=None,figsize:tuple=(4,6),
                       ticks_fontsize:int=12,labels_fontsize:int=12,ax=None,
                       legend:bool=False,legend_awargs={'ncol':1}):
    """
    Plot cell proportion of each cell type in each visual cluster.

    Arguments:
        adata: AnnData object.
        celltype_clusters: Cell type clusters.
        groupby: Visual clusters.
        groupby_li: Visual cluster list.
        figsize: Figure size.
        ticks_fontsize: Ticks fontsize.
        labels_fontsize: Labels fontsize.
        legend: Whether to show legend.
    
    
    """

    b=pd.DataFrame(columns=['cell_type','value','Week'])
    visual_clusters=groupby
    visual_li=groupby_li
    if visual_li==None:
        adata.obs[visual_clusters]=adata.obs[visual_clusters].astype('category')
        visual_li=adata.obs[visual_clusters].cat.categories
    
    for i in visual_li:
        b1=pd.DataFrame()
        test=adata.obs.loc[adata.obs[visual_clusters]==i,celltype_clusters].value_counts()
        b1['cell_type']=test.index
        b1['value']=test.values/test.sum()
        b1['Week']=i.replace('Retinoblastoma_','')
        b=pd.concat([b,b1])
    
    plt_data2=adata.obs[celltype_clusters].value_counts()
    plot_data2_color_dict=dict(zip(adata.obs[celltype_clusters].cat.categories,
                                   adata.uns['{}_colors'.format(celltype_clusters)]))
    plt_data3=adata.obs[visual_clusters].value_counts()
    plot_data3_color_dict=dict(zip([i.replace('Retinoblastoma_','') for i in adata.obs[visual_clusters].cat.categories],adata.uns['{}_colors'.format(visual_clusters)]))
    b['cell_type_color'] = b['cell_type'].map(plot_data2_color_dict)
    b['stage_color']=b['Week'].map(plot_data3_color_dict)
    
    if ax==None:
        fig, ax = plt.subplots(figsize=figsize)
    #用ax控制图片
    #sns.set_theme(style="whitegrid")
    #sns.set_theme(style="ticks")
    n=0
    all_celltype=adata.obs[celltype_clusters].cat.categories
    for i in all_celltype:
        if n==0:
            test1=b[b['cell_type']==i]
            ax.bar(x=test1['Week'],height=test1['value'],width=0.8,color=list(set(test1['cell_type_color']))[0], label=i)
            bottoms=test1['value'].values
        else:
            test2=b[b['cell_type']==i]
            ax.bar(x=test2['Week'],height=test2['value'],bottom=bottoms,width=0.8,color=list(set(test2['cell_type_color']))[0], label=i)
            test1=test2
            bottoms+=test1['value'].values
        n+=1
    if legend!=False:
        plt.legend(bbox_to_anchor=(1.05, -0.05), loc=3, borderaxespad=0,fontsize=10,**legend_awargs)
    
    plt.grid(False)
    
    plt.grid(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)

    # 设置左边和下边的坐标刻度为透明色
    #ax.yaxis.tick_left()
    #ax.xaxis.tick_bottom()
    #ax.xaxis.set_tick_params(color='none')
    #ax.yaxis.set_tick_params(color='none')

    # 设置左边和下边的坐标轴线为独立的线段
    ax.spines['left'].set_position(('outward', 10))
    ax.spines['bottom'].set_position(('outward', 10))

    plt.xticks(fontsize=ticks_fontsize,rotation=90)
    plt.yticks(fontsize=ticks_fontsize)
    plt.xlabel(groupby,fontsize=labels_fontsize)
    plt.ylabel('Cells per Stage',fontsize=labels_fontsize)
    #fig.tight_layout()
    if ax==None:
        return fig,ax
    


def embedding_celltype(adata:AnnData,figsize:tuple=(6,4),basis:str='umap',
                            celltype_key:str='major_celltype',title:str=None,
                            celltype_range:tuple=(2,9),
                            embedding_range:tuple=(3,10),
                            xlim:int=-1000)->tuple:
    """
    Plot embedding with celltype color by omicverse

    Arguments:
        adata: AnnData object  
        figsize: figure size
        basis: embedding method
        celltype_key: celltype key in adata.obs
        title: figure title
        celltype_range: celltype range to plot
        embedding_range: embedding range to plot
        xlim: x axis limit

    Returns:
        fig : figure and axis
        ax: axis
    
    """

    adata.obs[celltype_key]=adata.obs[celltype_key].astype('category')
    if pd.__version__>="2.0.0":
        cell_num_pd=pd.DataFrame(adata.obs[celltype_key].value_counts())
        cell_num_pd[celltype_key]=cell_num_pd['count']
    else:
        cell_num_pd=pd.DataFrame(adata.obs[celltype_key].value_counts())
        
    if '{}_colors'.format(celltype_key) in adata.uns.keys():
        cell_color_dict=dict(zip(adata.obs[celltype_key].cat.categories.tolist(),
                        adata.uns['{}_colors'.format(celltype_key)]))
    else:
        if len(adata.obs[celltype_key].cat.categories)>28:
            cell_color_dict=dict(zip(adata.obs[celltype_key].cat.categories,sc.pl.palettes.default_102))
        else:
            cell_color_dict=dict(zip(adata.obs[celltype_key].cat.categories,sc.pl.palettes.zeileis_28))

    if figsize==None:
        if len(adata.obs[celltype_key].cat.categories)<10:
            fig = plt.figure(figsize=(6,4))
        else:
            print('The number of cell types is too large, please set the figsize parameter')
            return
    else:
        fig = plt.figure(figsize=figsize)
    grid = plt.GridSpec(10, 10)
    ax1 = fig.add_subplot(grid[:, embedding_range[0]:embedding_range[1]])       # 占据第一行的所有列
    ax2 = fig.add_subplot(grid[celltype_range[0]:celltype_range[1], :2]) 
    # 定义子图的大小和位置
         # 占据第二行的前两列
    #ax3 = fig.add_subplot(grid[1:, 2])      # 占据第二行及以后的最后一列
    #ax4 = fig.add_subplot(grid[2, 0])       # 占据最后一行的第一列
    #ax5 = fig.add_subplot(grid[2, 1])       # 占据最后一行的第二列

    sc.pl.embedding(
        adata,
        basis=basis,
        color=[celltype_key],
        title='',
        frameon=False,
        #wspace=0.65,
        ncols=3,
        ax=ax1,
        legend_loc=False,
        show=False
    )



    for idx,cell in zip(range(cell_num_pd.shape[0]),
                        adata.obs[celltype_key].cat.categories):
        ax2.scatter(100,
                cell,c=cell_color_dict[cell],s=50)
        ax2.plot((100,cell_num_pd.loc[cell,celltype_key]),(idx,idx),
                c=cell_color_dict[cell],lw=4)
        ax2.text(100,idx+0.2,
                cell+'('+str("{:,}".format(cell_num_pd.loc[cell,celltype_key]))+')',fontsize=11)
    ax2.set_xlim(xlim,cell_num_pd.iloc[1].values[0]) 
    ax2.text(xlim,idx+1,title,fontsize=12)
    ax2.grid(False)
    #ax2.legend(bbox_to_anchor=(1.05, -0.05), loc=3, borderaxespad=0,fontsize=10,**legend_awargs)
    ax2.spines['top'].set_visible(False)
    ax2.spines['right'].set_visible(False)
    ax2.spines['bottom'].set_visible(False)
    ax2.spines['left'].set_visible(False)
    ax2.axis('off')

    # ——关键：确保 ax2 没有图例——
    if ax1.get_legend() is not None:   # 如果有，就移除
        ax1.get_legend().remove()
    if ax2.get_legend() is not None:   # 如果有，就移除
        ax2.get_legend().remove()

    return fig,[ax1,ax2]


def ConvexHull(adata:AnnData,basis:str,cluster_key:str,
                    hull_cluster:str,ax,color=None,alpha:float=0.2):
    """
    Plot the ConvexHull for a cluster in embedding

    Arguments:
        adata: AnnData object
        basis: embedding method in adata.obsm
        cluster_key: cluster key in adata.obs
        hull_cluster: cluster to plot for ConvexHull
        ax: axes
        color: color for ConvexHull
        alpha: alpha for ConvexHull

    Returns:
        ax: axes
    
    """
    from scipy.spatial import ConvexHull
    adata.obs[cluster_key]=adata.obs[cluster_key].astype('category')
    if '{}_colors'.format(cluster_key) in adata.uns.keys():
        print('{}_colors'.format(cluster_key))
        type_color_all=dict(zip(adata.obs[cluster_key].cat.categories,adata.uns['{}_colors'.format(cluster_key)]))
    else:
        if len(adata.obs[cluster_key].cat.categories)>28:
            type_color_all=dict(zip(adata.obs[cluster_key].cat.categories,sc.pl.palettes.default_102))
        else:
            type_color_all=dict(zip(adata.obs[cluster_key].cat.categories,sc.pl.palettes.zeileis_28))
    
    #color_dict=dict(zip(adata.obs[cluster_key].cat.categories,adata.uns[f'{cluster_key}_colors']))
    points=adata[adata.obs[cluster_key]==hull_cluster].obsm[basis]
    hull = ConvexHull(points)
    vert = np.append(hull.vertices, hull.vertices[0])  # close the polygon by appending the first point at the end
    if color==None:
        ax.plot(points[vert, 0], points[vert, 1], '--', c=type_color_all[hull_cluster])
        ax.fill(points[vert, 0], points[vert, 1], c=type_color_all[hull_cluster], alpha=alpha)
    else:
        ax.plot(points[vert, 0], points[vert, 1], '--', c=color)
        ax.fill(points[vert, 0], points[vert, 1], c=color, alpha=alpha)
    return ax


def embedding_adjust(
    adata, groupby, exclude=(), 
    basis='X_umap',ax=None, adjust_kwargs=None, text_kwargs=None
):
    """ 
    Get locations of cluster median . Borrowed from scanpy github forum.
    """
    if adjust_kwargs is None:
        adjust_kwargs = {"text_from_points": False}
    if text_kwargs is None:
        text_kwargs = {}

    medians = {}

    for g, g_idx in adata.obs.groupby(groupby).groups.items():
        if g in exclude:
            continue
        medians[g] = np.median(adata[g_idx].obsm[basis], axis=0)

    if ax is None:
        texts = [
            plt.text(x=x, y=y, s=k, **text_kwargs) for k, (x, y) in medians.items()
        ]
    else:
        texts = [ax.text(x=x, y=y, s=k, **text_kwargs) for k, (x, y) in medians.items()]
    from adjustText import adjust_text
    adjust_text(texts, **adjust_kwargs)


def embedding_density(adata,basis,groupby,target_clusters,**kwargs):
    if 'X_' in basis:
        basis1=basis.split('_')[1]
    sc.tl.embedding_density(adata,
                       basis=basis1,
                       groupby=groupby,
                       key_added='temp_density')
    adata.obs.loc[adata.obs[groupby]!=target_clusters,'temp_density']=0
    return embedding(adata,
                  basis=basis,
                  color=['temp_density'],
                    title=target_clusters,
                   **kwargs
                 )

def bardotplot(adata,groupby,color,figsize=(8,3),return_values=False,
               fontsize=12,xlabel='',ylabel='',xticks_rotation=90,ax=None,
               bar_kwargs=None,scatter_kwargs=None):
    
    if bar_kwargs is None:
        bar_kwargs = {}
    if scatter_kwargs is None:
        scatter_kwargs = {}
    
    var_ticks=False
    obs_ticks=False
    plot_text_=color
    if (plot_text_ in adata.var_names):
        adata1=adata
        var_ticks=True
    elif plot_text_ in adata.obs.columns:
        adata1=adata
        obs_ticks=True
    elif (adata.raw!=None) and (plot_text_ in adata.raw.var_names):
        adata1=adata1.raw.to_adata()
        var_ticks=True
    else:
        print(f'Please check the `{color}` key in adata.obs or adata.var')
        return
    adata1.obs[groupby]=adata1.obs[groupby].astype('category')
    
    if var_ticks==True:
        plot_data=pd.DataFrame()
        max_len=0
        for group in adata1.obs[groupby].cat.categories:
            if max_len<len(adata1[adata1.obs[groupby]==group,plot_text_].to_df().values.reshape(-1)):
                max_len=len(adata1[adata1.obs[groupby]==group,plot_text_].to_df().values.reshape(-1))
        for group in adata1.obs[groupby].cat.categories:
            t_data1=list(adata1[adata1.obs[groupby]==group,plot_text_].to_df().values.reshape(-1))
            while len(t_data1)<max_len:
                t_data1.append(np.nan)
            plot_data[group]=t_data1
    elif obs_ticks==True:
        plot_data=pd.DataFrame()
        max_len=0
        for group in adata1.obs[groupby].cat.categories:
            if max_len<len(adata1.obs.loc[adata1.obs[groupby]==group,plot_text_].values.reshape(-1)):
                max_len=len(adata1.obs.loc[adata1.obs[groupby]==group,plot_text_].values.reshape(-1))
        for group in adata1.obs[groupby].cat.categories:
            t_data1=list(adata1.obs.loc[adata1.obs[groupby]==group,plot_text_].values.reshape(-1))
            while len(t_data1)<max_len:
                t_data1.append(np.nan)
            plot_data[group]=t_data1
    
    if return_values==True:
        return plot_data
    
    
    if ax==None:
        fig, ax = plt.subplots(figsize=figsize)
        
    xbar = np.arange(len(plot_data.columns.to_numpy()))
    #color_list_dot=[ov.utils.green_color[0],ov.utils.green_color[1],ov.utils.red_color[0],'#EC9DC5','#5BC23D']
    #color_list_dot=adata.uns['clusters_colors']
    if '{}_colors'.format(groupby) in adata.uns.keys():
        color_list_dot=adata.uns['{}_colors'.format(groupby)]
    else:
        if len(adata.obs[groupby].cat.categories)>28:
            color_list_dot=sc.pl.palettes.default_102
        else:
            color_list_dot=sc.pl.palettes.zeileis_28
            
    plt.bar(x=plot_data.columns, 
            height=plot_data.describe().loc['mean'], 
            yerr=plot_data.sem(), 
            color=color_list_dot, 
            zorder=1, #fill=False,
            edgecolor=color_list_dot,
            error_kw={'elinewidth': None, 'capthick': None},**bar_kwargs)
    bw=0.4
    for cols in range(len(plot_data.columns.to_numpy())):
        # get markers from here https://matplotlib.org/3.1.1/api/markers_api.html
        plt.scatter(x=np.linspace(xbar[cols]-bw/2, xbar[cols]+bw/2, int(plot_data.describe().loc['count'][cols])),
                   y=plot_data[plot_data.columns[cols]].dropna(), 
                    color=color_list_dot[cols], zorder=1, 
                   **scatter_kwargs)


    plt.grid(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)
    ax.spines['left'].set_position(('outward', 10))
    ax.spines['bottom'].set_position(('outward', 10))

    plt.xticks(rotation=xticks_rotation,fontsize=fontsize)
    plt.xlabel(xlabel,fontsize=fontsize+1)
    plt.ylabel(ylabel,fontsize=fontsize+1)
    plt.title(plot_text_,fontsize=fontsize+1)
    if ax==None:
        return fig,ax
    
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from scipy.stats import kruskal

def single_group_boxplot(adata,
                         groupby: str = '',
                         color: str = '',
                         type_color_dict: dict = None,
                         title: str = '',
                         ylabel: str = '',
                         kruskal_test: bool = False,
                         figsize: tuple = (4, 4),
                         x_ticks_plot: bool = False,
                         legend_plot: bool = True,
                         bbox_to_anchor: tuple = (1, 0.55),
                         save: bool = False,
                         point_number: int = 5,
                         save_pathway: str = '',
                         sort: bool = True,
                         scatter_kwargs: dict = None,
                         ax = None,
                         fontsize = 12,
                        ):
    """
    adata (AnnData object): The data object containing the information for plotting.
    groupby (str): The variable used for grouping the data.
    color (str): The variable used for coloring the data points.
    type_color_dict (dict): A dictionary mapping group categories to specific colors.
    title (str): The title for the plot.
    ylabel (str): The label for the y-axis.
    kruskal_test (bool): Whether to perform a Kruskal-Wallis test and display the p-value on the plot.
    figsize (tuple): The size of the plot figure in inches (width, height).
    x_ticks_plot (bool): Whether to display x-axis tick labels.
    legend_plot (bool): Whether to display a legend for the groups.
    bbox_to_anchor (tuple): The position of the legend bbox (x, y) in axes coordinates.
    save (bool): Whether to save the plot to a file.
    point_number (int): The number of data points to be plotted for each group.
    save_pathway (str): The file path for saving the plot (if save is True).
    sort (bool): Whether to sort the groups based on their mean values.
    scatter_kwargs (dict): Additional keyword arguments for customizing the scatter plot.
    ax (matplotlib.axes.Axes): A pre-existing axes object for plotting (optional).
    
    Example:
    ov.pl.single_group_boxplot(adata,groupby='clusters',
             color='Sox_aucell',
             type_color_dict=dict(zip(pd.Categorical(adata.obs['clusters']).categories, adata.uns['clusters_colors'])),
             x_ticks_plot=True,
             figsize=(5,4),
             kruskal_test=True,
             ylabel='Sox_aucell',
             legend_plot=False,
             bbox_to_anchor=(1,1),
             title='Expression',
             scatter_kwargs={'alpha':0.8,'s':10,'marker':'o'},
             point_number=15,
             sort=False,
             save=False,
             )
    """

    if scatter_kwargs is None:
        scatter_kwargs = {}

    # Create an empty dictionary to store results
    plot_data = {}

    var_ticks = False
    obs_ticks = False
    plot_text_ = color
    if (plot_text_ in adata.var_names):
        adata1 = adata.copy()
        var_ticks = True
    elif plot_text_ in adata.obs.columns:
        adata1 = adata.copy()
        obs_ticks = True
    elif (adata.raw is not None) and (plot_text_ in adata.raw.var_names):
        adata1 = adata.raw.to_adata().copy()
        var_ticks = True
    else:
        print(f'Please check the `{color}` key in adata.obs or adata.var')
        return
    adata1.obs[groupby] = adata1.obs[groupby].astype('category')

    if var_ticks == True:
        adata1.obs[color] = adata1[:, plot_text_].to_df().values.flatten()
        #print(adata1.obs[color])

    # Categorize by groups
    
    for group in adata1.obs[groupby].cat.categories:
        plot_data[group] = np.array(adata1.obs.loc[adata1.obs[groupby] == group, color].tolist())

    if sort == True:
        sorted_keys = sorted(plot_data.keys(), key=lambda k: np.mean(plot_data[k]))
        sorted_plot_data = {key: plot_data[key] for key in sorted_keys}
        plot_data = sorted_plot_data

        sorted_colors = [type_color_dict[key] for key in sorted_keys]
        sc_color = sorted_colors
    else:
        sc_color = [type_color_dict[key] for key in plot_data.keys()]

    shake_dict = {}

    for group in adata1.obs[groupby].cat.categories:
        data_list = []
        gene_data = adata1.obs.loc[adata1.obs[groupby] == group, color].tolist()
        if len(gene_data) > point_number:
            bootstrap_data = np.random.choice(gene_data, size=point_number, replace=False)
        else:
            bootstrap_data = gene_data
        shake_dict[group] = np.array(bootstrap_data)

    if ax==None:
        fig, ax = plt.subplots(figsize=figsize)

    # Plot boxplots
    width = 0.8
    ticks = np.arange(len(plot_data))
    positions = np.arange(len(ticks))

    for num, (hue_data, hue_color) in enumerate(zip(plot_data.keys(), sc_color)):
        position = positions[num]
        b1 = ax.boxplot(plot_data[hue_data],
                        positions=[position],
                        sym='',
                        widths=width,
                        patch_artist=True
                        )
        plt.scatter(np.random.normal(position, 0.12, point_number),
                    shake_dict[hue_data],
                    c=hue_color, zorder=1, **scatter_kwargs)
        box = b1['boxes'][0]
        light_hue_color = tuple((min(1, c + 0.5 * (1 - c))) for c in plt.cm.colors.to_rgb(hue_color))
        box.set(facecolor=light_hue_color, edgecolor=hue_color, linewidth=2)
        plt.setp(b1['whiskers'], color=hue_color, linewidth=2)
        plt.setp(b1['caps'], color=hue_color, linewidth=2)
        plt.setp(b1['medians'], color=hue_color, linewidth=3)

    # Axis labels and title
    

    if x_ticks_plot == True:
        ax.set_xticks(positions)
        ax.set_xticklabels(plot_data.keys(), rotation=90, fontsize=fontsize)
    else:
        ax.set_xticklabels([])

    yticks = ax.get_yticks()
    ax.set_title(title, fontsize=fontsize+1,)
    plt.ylabel(ylabel, fontsize=fontsize+1, )
    plt.grid(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)
    ax.spines['left'].set_position(('outward', 10))
    ax.spines['bottom'].set_position(('outward', 10))

    if legend_plot == True:
        labels = list(plot_data.keys())
        patches = [mpatches.Patch(color=sc_color[i], label="{:s}".format(labels[i])) for i in range(len(plot_data))]
        ax.legend(handles=patches, bbox_to_anchor=bbox_to_anchor, ncol=1, fontsize=fontsize)

    if kruskal_test == True:
        data_list = [plot_data[key] for key in plot_data]
        statistic, p_value = kruskal(*data_list)

        if p_value < 0.0001:
            formatted_p_value = "{:.2e}".format(p_value)
        else:
            formatted_p_value = "{:.4f}".format(p_value)
        if p_value < 2.2e-16:
            formatted_p_value = 2.2e-16
            text = f"Kruskal-Wallis: P < {formatted_p_value}"
        else:
            text = f"Kruskal-Wallis: P = {formatted_p_value}"
        ax.text(0.05, 0.95, text, transform=ax.transAxes, fontsize=fontsize, fontweight='bold', verticalalignment='top',
                 bbox=dict(facecolor='white', edgecolor='white', boxstyle='round,pad=0.5'))

    if save == True:
        plt.savefig(save_pathway, dpi=300, bbox_inches='tight')

    if ax is None:
        return fig, ax

def contour(ax,adata,groupby,clusters,basis='X_umap',
            grid_density=100,contour_threshold=0.1,
           **kwargs):
    from scipy.stats import gaussian_kde
    umap_embedding=adata[adata.obs[groupby].isin(clusters)].obsm[basis]
    kde = gaussian_kde(umap_embedding.T)
    # 生成网格
    x_grid = np.linspace(min(umap_embedding[:, 0]), max(umap_embedding[:, 0]), grid_density)
    y_grid = np.linspace(min(umap_embedding[:, 1]), max(umap_embedding[:, 1]), grid_density)
    X_grid, Y_grid = np.meshgrid(x_grid, y_grid)
    positions = np.vstack([X_grid.ravel(), Y_grid.ravel()])
    Z = np.reshape(kde(positions).T, X_grid.shape)
    
    # 找到最外层轮廓的阈值
    threshold = np.max(Z) * contour_threshold  # 通过调整阈值的倍数来控制轮廓的密度

    # 绘制最外层轮廓
    contour = ax.contour(X_grid, Y_grid, Z, 
                         levels=[threshold],
                        **kwargs)

    # 获取最外层轮廓的路径
    #outer_contour_path = contour.collections[0].get_paths()[0]
    #outer_contour_x, outer_contour_y = outer_contour_path.vertices[:, 0], outer_contour_path.vertices[:, 1]

    return ax


def plot_boxplots(  # pragma: no cover
        data,
        feature_name: str,
        modality_key: str = "coda",
        y_scale = "relative",
        plot_facets: bool = False,
        add_dots: bool = False,
        cell_types=None,
        args_boxplot=None,
        args_swarmplot= None,
        palette= "Blues",
        show_legend= True,
        level_order=None,
        figsize  = None,
        dpi= 100,
        return_fig= None,
        ax = None,
        show = None,
        save = None,
    ):
        """Grouped boxplot visualization.

         The cell counts for each cell type are shown as a group of boxplots
         with intra--group separation by a covariate from data.obs.

        Args:
            data: AnnData object or MuData object
            feature_name: The name of the feature in data.obs to plot
            modality_key: If data is a MuData object, specify which modality to use.
            y_scale: Transformation to of cell counts. Options: "relative" - Relative abundance, "log" - log(count),
                     "log10" - log10(count), "count" - absolute abundance (cell counts).
            plot_facets: If False, plot cell types on the x-axis. If True, plot as facets.
            add_dots: If True, overlay a scatterplot with one dot for each data point.
            cell_types: Subset of cell types that should be plotted.
            args_boxplot: Arguments passed to sns.boxplot.
            args_swarmplot: Arguments passed to sns.swarmplot.
            figsize: Figure size.
            dpi: Dpi setting.
            palette: The seaborn color map for the barplot.
            show_legend: If True, adds a legend.
            level_order: Custom ordering of bars on the x-axis.

        Returns:
            Depending on `plot_facets`, returns a :class:`~matplotlib.axes.Axes` (`plot_facets = False`)
            or :class:`~sns.axisgrid.FacetGrid` (`plot_facets = True`) object

        Examples:
            >>> import pertpy as pt
            >>> haber_cells = pt.dt.haber_2017_regions()
            >>> sccoda = pt.tl.Sccoda()
            >>> mdata = sccoda.load(haber_cells, type="cell_level", generate_sample_level=True, cell_type_identifier="cell_label", \
                sample_identifier="batch", covariate_obs=["condition"])
            >>> sccoda.plot_boxplots(mdata, feature_name="condition", add_dots=True)

        Preview:
            .. image:: /_static/docstring_previews/sccoda_boxplots.png
        """
        if args_boxplot is None:
            args_boxplot = {}
        if args_swarmplot is None:
            args_swarmplot = {}
        #if isinstance(data, MuData):
        #    data = data[modality_key]
        #if isinstance(data, AnnData):
        #    data = data
        # y scale transformations
        if y_scale == "relative":
            sample_sums = np.sum(data.X, axis=1, keepdims=True)
            X = data.X / sample_sums
            value_name = "Proportion"
        # add pseudocount 0.5 if using log scale
        elif y_scale == "log":
            X = data.X.copy()
            X[X == 0] = 0.5
            X = np.log(X)
            value_name = "log(count)"
        elif y_scale == "log10":
            X = data.X.copy()
            X[X == 0] = 0.5
            X = np.log(X)
            value_name = "log10(count)"
        elif y_scale == "count":
            X = data.X
            value_name = "count"
        else:
            raise ValueError("Invalid y_scale transformation")

        count_df = pd.DataFrame(X, columns=data.var.index, index=data.obs.index).merge(
            data.obs[feature_name], left_index=True, right_index=True
        )
        plot_df = pd.melt(count_df, id_vars=feature_name, var_name="Cell type", value_name=value_name)
        if cell_types is not None:
            plot_df = plot_df[plot_df["Cell type"].isin(cell_types)]

        # Currently disabled because the latest statsannotations does not support the latest seaborn.
        # We had to drop the dependency.
        # Get credible effects results from model
        # if draw_effects:
        #     if model is not None:
        #         credible_effects_df = model.credible_effects(data, modality_key).to_frame().reset_index()
        #     else:
        #         print("[bold yellow]Specify a tasCODA model to draw effects")
        #     credible_effects_df[feature_name] = credible_effects_df["Covariate"].str.removeprefix(f"{feature_name}[T.")
        #     credible_effects_df[feature_name] = credible_effects_df[feature_name].str.removesuffix("]")
        #     credible_effects_df = credible_effects_df[credible_effects_df["Final Parameter"]]

        # If plot as facets, create a FacetGrid and map boxplot to it.
        if plot_facets:
            if level_order is None:
                level_order = pd.unique(plot_df[feature_name])

            K = X.shape[1]

            if figsize is not None:
                height = figsize[0]
                aspect = np.round(figsize[1] / figsize[0], 2)
            else:
                height = 3
                aspect = 2

            g = sns.FacetGrid(
                plot_df,
                col="Cell type",
                sharey=False,
                col_wrap=int(np.floor(np.sqrt(K))),
                height=height,
                aspect=aspect,
            )
            g.map(
                sns.boxplot,
                feature_name,
                value_name,
                palette=palette,
                order=level_order,
                **args_boxplot,
            )

            if add_dots:
                if "hue" in args_swarmplot:
                    hue = args_swarmplot.pop("hue")
                else:
                    hue = None

                if hue is None:
                    g.map(
                        sns.swarmplot,
                        feature_name,
                        value_name,
                        color="black",
                        order=level_order,
                        **args_swarmplot,
                    ).set_titles("{col_name}")
                else:
                    g.map(
                        sns.swarmplot,
                        feature_name,
                        value_name,
                        hue,
                        order=level_order,
                        **args_swarmplot,
                    ).set_titles("{col_name}")

            if save:
                plt.savefig(save, bbox_inches="tight")
            if show:
                plt.show()
            if return_fig:
                return plt.gcf()
            if not (show or save):
                return g
            return None

        # If not plot as facets, call boxplot to plot cell types on the x-axis.
        else:
            if level_order:
                args_boxplot["hue_order"] = level_order
                args_swarmplot["hue_order"] = level_order

            _, ax = plt.subplots(figsize=figsize, dpi=dpi)

            ax = sns.boxplot(
                x="Cell type",
                y=value_name,
                hue=feature_name,
                data=plot_df,
                fliersize=1,
                palette=palette,
                ax=ax,
                **args_boxplot,
            )

            # Currently disabled because the latest statsannotations does not support the latest seaborn.
            # We had to drop the dependency.
            # if draw_effects:
            #     pairs = [
            #         [(row["Cell Type"], row[feature_name]), (row["Cell Type"], "Control")]
            #         for _, row in credible_effects_df.iterrows()
            #     ]
            #     annot = Annotator(ax, pairs, data=plot_df, x="Cell type", y=value_name, hue=feature_name)
            #     annot.configure(test=None, loc="outside", color="red", line_height=0, verbose=False)
            #     annot.set_custom_annotations([row[feature_name] for _, row in credible_effects_df.iterrows()])
            #     annot.annotate()

            if add_dots:
                sns.swarmplot(
                    x="Cell type",
                    y=value_name,
                    data=plot_df,
                    hue=feature_name,
                    ax=ax,
                    dodge=True,
                    palette="dark:black",
                    **args_swarmplot,
                )

            cell_types = pd.unique(plot_df["Cell type"])
            ax.set_xticklabels(cell_types, rotation=90)

            if show_legend:
                handles, labels = ax.get_legend_handles_labels()
                handout = []
                labelout = []
                for h, l in zip(handles, labels, strict=False):
                    if l not in labelout:
                        labelout.append(l)
                        handout.append(h)
                ax.legend(
                    handout,
                    labelout,
                    loc="upper left",
                    bbox_to_anchor=(1, 1),
                    ncol=1,
                    title=feature_name,
                )

            if save:
                plt.savefig(save, bbox_inches="tight")
            if show:
                plt.show()
            if return_fig:
                return plt.gcf()
            if not (show or save):
                return ax
            return None
        

def cellstackarea(adata,celltype_clusters:str,groupby:str,
                       groupby_li=None,figsize:tuple=(4,6),
                       ticks_fontsize:int=12,labels_fontsize:int=12,ax=None,
                       legend:bool=False,legend_awargs={},text_show=False,):
    """
    Plot the cell type percentage in each groupby category
    
    """
    df = adata.obs[[groupby, celltype_clusters]]

    # 计算每个样本类型中每个细胞类型的数量
    count_df = df.groupby([groupby, celltype_clusters]).size().reset_index(name='count')
    
    # 计算每个样本类型中的总数
    total_count_df = count_df.groupby(groupby)['count'].sum().reset_index(name='total_count')
    
    # 将总数合并回原数据框
    count_df = count_df.merge(total_count_df, on=groupby)
    
    # 计算百分比
    count_df['percentage'] = count_df['count'] / count_df['total_count'] * 100
    
    # 将数据从长格式转换为宽格式，以便绘制面积图
    pivot_df = count_df.pivot(index=groupby, columns=celltype_clusters, values='percentage').fillna(0)
    if groupby_li!=None:
        pivot_df=pivot_df.loc[groupby_li]

    
    # 使用 matplotlib 绘制面积图
    if ax==None:
        fig, ax = plt.subplots(figsize=figsize)
    
    # 为每种细胞类型绘制面积图
    cell_types = pivot_df.columns
    bottom = pd.Series([0] * len(pivot_df), index=pivot_df.index)

    adata.obs[celltype_clusters]=adata.obs[celltype_clusters].astype('category')
    if '{}_colors'.format(celltype_clusters) in adata.uns.keys():
        print('{}_colors'.format(celltype_clusters))
        type_color_all=dict(zip(adata.obs[celltype_clusters].cat.categories,adata.uns['{}_colors'.format(celltype_clusters)]))
    else:
        if len(adata.obs[celltype_clusters].cat.categories)>28:
            type_color_all=dict(zip(adata.obs[celltype_clusters].cat.categories,sc.pl.palettes.default_102))
        else:
            type_color_all=dict(zip(adata.obs[celltype_clusters].cat.categories,sc.pl.palettes.zeileis_28))
    
    
    
    for cell_type in cell_types:
        ax.fill_between(pivot_df.index, bottom, bottom + pivot_df[cell_type], label=cell_type,
                       color=type_color_all[cell_type])
        max_index = pivot_df[cell_type].idxmax()
        if text_show==True:
            ax.text(max_index,bottom[max_index]+pivot_df.loc[max_index,cell_type]/2,cell_type,fontsize=ticks_fontsize-1)
        
        bottom += pivot_df[cell_type]
    
    if legend!=False:
        plt.legend(bbox_to_anchor=(1.05, -0.05), loc=3, borderaxespad=0,
                   fontsize=labels_fontsize,**legend_awargs)
    
    plt.grid(False)
    
    plt.grid(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)

    # 设置左边和下边的坐标刻度为透明色
    #ax.yaxis.tick_left()
    #ax.xaxis.tick_bottom()
    #ax.xaxis.set_tick_params(color='none')
    #ax.yaxis.set_tick_params(color='none')

    # 设置左边和下边的坐标轴线为独立的线段
    ax.spines['left'].set_position(('outward', 10))
    ax.spines['bottom'].set_position(('outward', 10))

    plt.xticks(fontsize=ticks_fontsize,rotation=90)
    plt.yticks(fontsize=ticks_fontsize)
    plt.xlabel(groupby,fontsize=labels_fontsize)
    plt.ylabel('Cells per Stage',fontsize=labels_fontsize)
    #fig.tight_layout()
    if ax==None:
        return fig,ax



def violin(adata,keys=None,groupby=None,ax=None,figsize=(4,4),fontsize=12,
           ticks_fontsize=None,rotation=90,**kwargs):
    if ax==None:
        fig, ax = plt.subplots(figsize=figsize)
    sc.pl.violin(adata,keys=keys,groupby=groupby,ax=ax,show=False,**kwargs)
    plt.grid(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)
    ax.spines['left'].set_position(('outward', 10))
    ax.spines['bottom'].set_position(('outward', 10))
    if ticks_fontsize==None:
        ticks_fontsize=fontsize-1

    plt.xticks(fontsize=ticks_fontsize,rotation=rotation)
    plt.yticks(fontsize=ticks_fontsize)
    plt.xlabel(groupby,fontsize=fontsize)
    plt.ylabel(keys,fontsize=fontsize)

    if ax==None:
        return fig,ax

    #plt.xticks(fontsize=ticks_fontsize,rotation=90)
    #plt.yticks(fontsize=ticks_fontsize)

def violin_box(adata, keys, groupby, ax=None, figsize=(4,4), show=True, max_strip_points=1000):
    import colorcet
    from scipy.sparse import issparse  
    
    # 获取 y 数据
    y = None
    if not adata.raw is None and keys in adata.raw.var_names:
        y = adata.raw[:, keys].X
    elif keys in adata.obs.columns:
        y = adata.obs[keys].values
    elif keys in adata.var_names:
        y = adata[:, keys].X
    else:
        raise ValueError(f'{keys} not found in adata.raw.var_names, adata.var_names, or adata.obs.columns')
    
    if issparse(y):
        y = y.toarray().reshape(-1)
    else:
        y = y.reshape(-1)
    
    # 获取 x 数据
    x = adata.obs[groupby].values.reshape(-1)
    
    # 创建绘图数据
    plot_data = pd.DataFrame({groupby: x, keys: y})
    
    # 创建图形和轴
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    
    # 获取或设置颜色
    from ._palette import palette_28,palette_56,palette_112,sc_color
    if f'{groupby}_colors' not in adata.uns or adata.uns[f'{groupby}_colors'] is None:
        #colors = ['#%02x%02x%02x' % tuple([int(k * 255) for k in i]) for i in colorcet.glasbey_bw_minc_20_maxl_70]
        if len(adata.obs[groupby].unique())>56:
            colors=palette_112
        elif len(adata.obs[groupby].unique())>28:
            colors=palette_56
        else:
            colors=sc_color
        adata.uns[f'{groupby}_colors'] = colors[:len(adata.obs[groupby].unique())]
    
    # 绘制小提琴图
    sns.violinplot(x=groupby, y=keys, data=plot_data, hue=groupby, dodge=False,
                   palette=adata.uns[f'{groupby}_colors'], scale="width", inner=None, ax=ax,
                   legend=False)
    
    # 调整小提琴图
    xlim, ylim = ax.get_xlim(), ax.get_ylim()
    for violin in ax.collections:
        bbox = violin.get_paths()[0].get_extents()
        x0, y0, width, height = bbox.bounds
        violin.set_clip_path(plt.Rectangle((x0, y0), width / 2, height, transform=ax.transData))
    
    
    
    # 限制 stripplot 的数据点数量
    if len(plot_data) > max_strip_points:
        plot_data = plot_data.sample(max_strip_points)
    
    # 绘制 stripplot
    old_len_collections = len(ax.collections)
    sns.stripplot(x=groupby, y=keys, data=plot_data, hue=groupby,
                  palette=adata.uns[f'{groupby}_colors'], dodge=False, ax=ax,
                  )
    
    # 调整 stripplot 点的位置
    for dots in ax.collections[old_len_collections:]:
        dots.set_offsets(dots.get_offsets() + np.array([0.12, 0]))

    # 绘制箱线图
    sns.boxplot(x=groupby, y=keys, data=plot_data, saturation=1, showfliers=False,
                width=0.3, boxprops={'zorder': 3, 'facecolor': 'none'}, ax=ax,
                )
    
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    #ax.set_xticklabels(ax.get_xticklabels(),rotation=90)
    #remove legend
    if ax.get_legend() is not None:
        ax.get_legend().remove()
    #ax.legend().set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)
    ax.spines['left'].set_position(('outward', 10))
    ax.spines['bottom'].set_position(('outward', 10))
    #

    if show:
        plt.show()
    
    return ax


import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import matplotlib.gridspec as gridspec

def dotplot_doublegroup(adata, gene, group1, group2, cmap='Reds',
                       standard_scale = 'group',figsize=(6, 4),layer=None):
    # 检查输入
    if gene not in adata.var_names:
        raise ValueError(f"Gene '{gene}' is not in the provided AnnData object.")
        
    if group1 not in adata.obs.columns:
        raise ValueError(f"Group '{group1}' is not in the provided AnnData object.")
        
    if group2 not in adata.obs.columns:
        raise ValueError(f"Group '{group2}' is not in the provided AnnData object.")
    
    # 将分组列转换为类别类型
    adata.obs[group1] = adata.obs[group1].astype('category')
    adata.obs[group2] = adata.obs[group2].astype('category')
    group1s = adata.obs[group1].cat.categories
    group2s = adata.obs[group2].cat.categories

    group_li_exp_mean_pd = pd.DataFrame(np.zeros((len(group1s), len(group2s))), index=group1s, columns=group2s)
    group_li_exp_size_pd = pd.DataFrame(np.zeros((len(group1s), len(group2s))), index=group1s, columns=group2s)

    for group1_ in group1s:
        adata1 = adata[adata.obs[group1] == group1_, [gene]]
        for group2_ in group2s:
            if layer is None:
                exp = adata1[adata1.obs[group2] == group2_, [gene]].to_df().values.reshape(-1)
            elif layer in adata1.layers.keys():
                exp = adata1[adata1.obs[group2] == group2_, [gene]].layers[layer].to_df().values.reshape(-1)
            else:
                raise ValueError(f"Layer '{layer}' is not in the provided AnnData object.")
            exp_larger_zero = exp[exp > 0]
            if len(exp) != 0:
                group_li_exp_size_pd.loc[group1_, group2_] = len(exp_larger_zero) / len(exp)
                group_li_exp_mean_pd.loc[group1_, group2_] = np.mean(exp)
            else:
                group_li_exp_size_pd.loc[group1_, group2_] = 0

    dot_color_df = group_li_exp_mean_pd
    
    if standard_scale == "group":
        dot_color_df = dot_color_df.sub(dot_color_df.min(1), axis=0)
        dot_color_df = dot_color_df.div(dot_color_df.max(1), axis=0).fillna(0)
    elif standard_scale == "var":
        dot_color_df -= dot_color_df.min(0)
        dot_color_df = (dot_color_df / dot_color_df.max(0)).fillna(0)
    else:
        pass
    
    # 设置常量
    dot_min = 0
    dot_max = 1
    size_exponent = 1
    largest_dot = 100
    smallest_dot = 10
    dot_edge_lw = 1
    size_title = 'Fraction of cells\nin group (%)'

    # 创建图形和网格布局
    fig = plt.figure(figsize=figsize)
    gs = gridspec.GridSpec(1, 2, width_ratios=[3, 1])
    ax = fig.add_subplot(gs[0])

    # 绘制圆点
    for y, (label, row_mean) in enumerate(dot_color_df.iterrows()):
        for x, (column, value_mean) in enumerate(row_mean.items()):
            value_size = group_li_exp_size_pd.loc[label, column]
            size = value_size * 500  # 调整大小
            color = plt.get_cmap(cmap)(value_mean / dot_color_df.values.max())
            ax.scatter(x, y, s=size, color=color, alpha=1, 
                       edgecolors='w', linewidth=0.5)

    # 设置轴
    ax.set_xticks(range(len(group_li_exp_mean_pd.columns)))
    ax.set_xticklabels(group_li_exp_mean_pd.columns, rotation=45)
    ax.set_yticks(range(len(group_li_exp_mean_pd.index)))
    ax.set_yticklabels(group_li_exp_mean_pd.index)
    ax.set_title('Dot Plot')

    # 添加颜色图例
    ax1 = fig.add_subplot(gs[1])
    legend_gs = ax1.get_subplotspec().subgridspec(10, 1)

    color_legend_ax = fig.add_subplot(legend_gs[7])
    mappable = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=group_li_exp_mean_pd.values.min(),
                                                                   vmax=group_li_exp_mean_pd.values.max()))
    plt.colorbar(mappable, cax=color_legend_ax, orientation="horizontal")
    color_legend_ax.set_title('Mean expression\nin group', fontsize="small")
    color_legend_ax.xaxis.set_tick_params(labelsize="small")

    # 添加点大小图例
    size_legend_ax = fig.add_subplot(legend_gs[2:5])
    diff = dot_max - dot_min
    step = 0.1 if 0.3 < diff <= 0.6 else 0.05 if diff <= 0.3 else 0.2

    size_range = np.arange(dot_max, dot_min, step * -1)[::-1]
    if dot_min != 0 or dot_max != 1:
        dot_range = dot_max - dot_min
        size_values = (size_range - dot_min) / dot_range
    else:
        size_values = size_range

    size = size_values ** size_exponent
    size = size * (largest_dot - smallest_dot) + smallest_dot

    size_legend_ax.scatter(
        np.arange(len(size)) + 0.5,
        np.repeat(0, len(size)),
        s=size * 5,
        color="gray",
        edgecolor="black",
        linewidth=dot_edge_lw,
        zorder=100,
    )
    size_legend_ax.set_xticks(np.arange(len(size)) + 0.5)
    labels = [f"{np.round((x * 100), decimals=0).astype(int)}" for x in size_range]
    size_legend_ax.set_xticklabels(labels, fontsize="small")

    size_legend_ax.tick_params(
        axis="y", left=False, labelleft=False, labelright=False
    )

    size_legend_ax.spines["right"].set_visible(False)
    size_legend_ax.spines["top"].set_visible(False)
    size_legend_ax.spines["left"].set_visible(False)
    size_legend_ax.spines["bottom"].set_visible(False)
    size_legend_ax.grid(visible=False)

    ymax = size_legend_ax.get_ylim()[1]
    size_legend_ax.set_ylim(-1.05 - largest_dot * 0.003, 4)
    size_legend_ax.set_title(size_title, y=ymax + 0.45, size="small")

    xmin, xmax = size_legend_ax.get_xlim()
    size_legend_ax.set_xlim(xmin - 0.15, xmax + 0.5)

    ax.grid(False)
    ax1.grid(False)
    ax1.axis(False)
    plt.show()


import torch
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from .model.aug import random_aug
from .utils import coords2adjacentmat
from timeit import default_timer as timer
from collections import OrderedDict
from tqdm import trange

def train_seq(graphs, args, dump_epoch_list, out_prefix, model):
    """The CAST MARK training function

    Args:
        graphs (List[Tuple(str, dgl.Graph, torch.Tensor)]): List of 3-member tuples, each tuple represents one tissue sample, containing sample name, a DGL graph object, and a feature matrix in the torch.Tensor format
        args (model_GCNII.Args): the Args object contains training parameters
        dump_epoch_list (List): A list of epoch id you hope training snapshots to be dumped, for debug use, empty by default
        out_prefix (str): file name prefix for the snapshot files
        model (model_GCNII.CCA_SSG): the GNN model

    Returns:
        Tuple(Dict, List, CCA_SSG): returns a 3-member tuple, a dictionary containing the graph embeddings for each sample, a list of every loss value, and the trained model object
    """    
    model = model.to(args.device)

    optimizer = torch.optim.Adam(model.parameters(), lr=args.lr1, weight_decay=args.wd1)

    loss_log = []
    time_now = timer()
    
    t = trange(args.epochs, desc='', leave=True)
    for epoch in t:

        with torch.no_grad():
            if epoch in dump_epoch_list:
                model.eval()
                dump_embedding = OrderedDict()
                for name, graph, feat in graphs:
                    # graph = graph.to(args.device)
                    # feat = feat.to(args.device)
                    dump_embedding[name] = model.get_embedding(graph, feat)
                torch.save(dump_embedding, f'{out_prefix}_embed_dict_epoch{epoch}.pt')
                torch.save(loss_log, f'{out_prefix}_loss_log_epoch{epoch}.pt')
                print(f"Successfully dumped epoch {epoch}")

        losses = dict()
        model.train()
        optimizer.zero_grad()
        # print(f'Epoch: {epoch}')

        for name_, graph_, feat_ in graphs:
            with torch.no_grad():
                N = graph_.number_of_nodes()
                graph1, feat1 = random_aug(graph_, feat_, args.dfr, args.der)
                graph2, feat2 = random_aug(graph_, feat_, args.dfr, args.der)

                graph1 = graph1.add_self_loop()
                graph2 = graph2.add_self_loop()

            z1, z2 = model(graph1, feat1, graph2, feat2)

            c = torch.mm(z1.T, z2)
            c1 = torch.mm(z1.T, z1)
            c2 = torch.mm(z2.T, z2)

            c = c / N
            c1 = c1 / N
            c2 = c2 / N

            loss_inv = - torch.diagonal(c).sum()
            iden = torch.eye(c.size(0), device=args.device)
            loss_dec1 = (iden - c1).pow(2).sum()
            loss_dec2 = (iden - c2).pow(2).sum()
            loss = loss_inv + args.lambd * (loss_dec1 + loss_dec2)
            loss.backward()
            optimizer.step()
            
        # del graph1, feat1, graph2, feat2        
        loss_log.append(loss.item())
        time_step = timer() - time_now
        time_now += time_step
        # print(f'Loss: {loss.item()} step time={time_step:.3f}s')
        t.set_description(f'Loss: {loss.item():.3f} step time={time_step:.3f}s')
        t.refresh()
    
    model.eval()
    with torch.no_grad():
        dump_embedding = OrderedDict()
        for name, graph, feat in graphs:
            dump_embedding[name] = model.get_embedding(graph, feat)
    return dump_embedding, loss_log, model

# graph construction tools
def delaunay_dgl(sample_name, df, output_path,if_plot=True,strategy_t = 'convex'):
    coords = np.column_stack((np.array(df)[:,0],np.array(df)[:,1]))
    delaunay_graph = coords2adjacentmat(coords,output_mode = 'raw',strategy_t = strategy_t)
    if if_plot:
        positions = dict(zip(delaunay_graph.nodes, coords[delaunay_graph.nodes,:]))
        _, ax = plt.subplots(nrows=1, ncols=1, figsize=(10, 10))
        nx.draw(
            delaunay_graph,
            positions,
            ax=ax,
            node_size=1,
            node_color="#000000",
            edge_color="#5A98AF",
            alpha=0.6,
        )
        plt.axis('equal')
        plt.savefig(f'{output_path}/delaunay_{sample_name}.png')
    import dgl
    return dgl.from_networkx(delaunay_graph)


"""Module providing a encapsulation of spatrio."""
from typing import Any
import pandas as pd
import numpy as np
import scanpy as sc
from ..externel.spatrio.spatrio import ot_alignment,assign_coord

class CellMap(object):
    """Class representing the object of CellMap."""
    def __init__(self,
                    adata_sc,
                    adata_sp,
                    use_rep_sc='X_pca',
                    use_rep_sp='X_pca',
                 ) -> None:
        self.adata_sc=adata_sc
        self.adata_sp=adata_sp
        self.use_rep_sc=use_rep_sc
        self.use_rep_sp=use_rep_sp
        self.spatrio_decon=None
        self.spatrio_map=None
        self.adata_sp.obsm['spatial']=pd.DataFrame(np.array(self.adata_sp.obsm['spatial']),
                                    columns=['x','y'],index=self.adata_sp.obs.index)
        self.adata_sc.obsm['reduction']=pd.DataFrame(self.adata_sc.obsm[self.use_rep_sc],
                                    index=self.adata_sc.obs.index)

    def map(self,
            sc_type: str = 'celltype',
            sp_type: str = 'leiden',
            alpha: float = 0.1,
            aware_power: int = 2,
            resolution: int = 1,
            aware_spatial: bool = True,
            aware_multi: bool = True,
            use_gpu: bool = True,
            **kwargs: Any
        ) -> pd.DataFrame:

        ##spatial type
        if sp_type=='leiden' and 'leiden' not in self.adata_sp.obs.columns:
            sc.pp.neighbors(self.adata_sp,n_neighbors=15,use_rep=self.use_rep_sp)
            sc.tl.leiden(self.adata_sp,resolution=resolution)
            self.adata_sp.obs['type']=self.adata_sp.obs[sp_type].tolist()
        elif sp_type=='louvain' and 'louvain' not in self.adata_sp.obs.columns:
            sc.pp.neighbors(self.adata_sp,n_neighbors=15,use_rep=self.use_rep_sp)
            sc.tl.louvain(self.adata_sp)
            self.adata_sp.obs['type']=self.adata_sp.obs[sp_type].tolist()
        elif sp_type is None:
            aware_spatial=False
        else:
            self.adata_sp.obs['type']=self.adata_sp.obs[sp_type].tolist()

        ##single cell type
        if sc_type is None:
            aware_multi=False
        else:
            self.adata_sc.obs['type']=self.adata_sc.obs[sc_type].tolist()

        spatrio_decon = ot_alignment(adata1 = self.adata_sp, adata2 = self.adata_sc,
                                     alpha = alpha, aware_power = aware_power,
                aware_spatial = aware_spatial,
                aware_multi=aware_multi,use_gpu=use_gpu,**kwargs)
        self.spatrio_decon=spatrio_decon
        return spatrio_decon

    def assign_coord(self,**kwargs):
        """
        Assign spatial coordinates to single cell data.
        """
        spatrio_map = assign_coord(adata1 = self.adata_sp,adata2 = self.adata_sc,
                                   out_data = self.spatrio_decon,**kwargs)
        #self.adata_sp.obs=self.adata_sp.obs.join(spatrio_map.set_index('cell'))

        self.spatrio_map=spatrio_map

        loc1=pd.DataFrame(spatrio_map[['Cell_xcoord','Cell_ycoord','spot','spot_type',
                                       'value']].values,
                          columns=['Cell_xcoord','Cell_ycoord','spot','spot_type','value'],
                          index=spatrio_map['cell'].tolist())
        print('...assigning spatial coordinates to single cell data')
        adata_sc_copy=self.adata_sc.copy()
        adata_sc_copy=adata_sc_copy[loc1.index.tolist()]
        adata_sc_copy.obs['Cell_xcoord']=loc1.loc[adata_sc_copy.obs.index,'Cell_xcoord'].tolist()
        adata_sc_copy.obs['Cell_ycoord']=loc1.loc[adata_sc_copy.obs.index,'Cell_ycoord'].tolist()
        adata_sc_copy.obs['spot_type']=loc1.loc[adata_sc_copy.obs.index,'spot_type'].tolist()
        adata_sc_copy.obs['spot']=loc1.loc[adata_sc_copy.obs.index,'spot'].tolist()
        adata_sc_copy.obs['spot_value']=loc1.loc[adata_sc_copy.obs.index,'value'].tolist()

        print('...adding spatial coordinates to single cell data')
        adata_sc_copy.obsm['spatial']=loc1.loc[adata_sc_copy.obs.index,
                                               ['Cell_xcoord','Cell_ycoord']].values
        adata_sc_copy.uns['spatial']=self.adata_sp.uns['spatial'].copy()
        return adata_sc_copy
    # End-of-file (EOF)


class CellLoc(object):
    """Class representing the object of CellMap."""
    def __init__(self,
                    adata_sc,
                    adata_sp,
                    use_rep_sc='X_pca',
                    use_rep_sp='X_pca',
                 ) -> None:
        self.adata_sc=adata_sc
        self.adata_sp=adata_sp
        self.use_rep_sc=use_rep_sc
        self.use_rep_sp=use_rep_sp
        self.spatrio_decon=None
        self.spatrio_map=None
        self.adata_sp.obsm['spatial']=pd.DataFrame(np.array(self.adata_sp.obsm['spatial']),
                                    columns=['x','y'],index=self.adata_sp.obs.index)
        self.adata_sc.obsm['reduction']=pd.DataFrame(self.adata_sc.obsm[self.use_rep_sc],
                                    index=self.adata_sc.obs.index)

    def loc_map(self,
            sc_type: str = 'celltype',
            sp_type: str = 'leiden',
            alpha: float = 0.1,
            aware_power: int = 2,
            resolution: int = 1,
            aware_spatial: bool = True,
            aware_multi: bool = True,
            use_gpu: bool = True,
            **kwargs: Any
        ) -> pd.DataFrame:

        ##spatial type
        if sp_type=='leiden' and 'leiden' not in self.adata_sp.obs.columns:
            sc.pp.neighbors(self.adata_sp,n_neighbors=15,use_rep=self.use_rep_sp)
            sc.tl.leiden(self.adata_sp,resolution=resolution)
            self.adata_sp.obs['type']=self.adata_sp.obs[sp_type].tolist()
        elif sp_type=='louvain' and 'louvain' not in self.adata_sp.obs.columns:
            sc.pp.neighbors(self.adata_sp,n_neighbors=15,use_rep=self.use_rep_sp)
            sc.tl.louvain(self.adata_sp)
            self.adata_sp.obs['type']=self.adata_sp.obs[sp_type].tolist()
        elif sp_type is None:
            aware_spatial=False
        else:
            self.adata_sp.obs['type']=self.adata_sp.obs[sp_type].tolist()

        ##single cell type
        if sc_type is None:
            aware_multi=False
        else:
            self.adata_sc.obs['type']=self.adata_sc.obs[sc_type].tolist()

        spatrio_decon = ot_alignment(adata1 = self.adata_sp, adata2 = self.adata_sc,
                                     alpha = alpha, aware_power = aware_power,
                aware_spatial = aware_spatial,
                aware_multi=aware_multi,use_gpu=use_gpu,**kwargs)
        self.spatrio_decon=spatrio_decon
        return spatrio_decon

    def load_map(self,map_info,
                 sc_type: str = 'celltype',
                 sp_type: str = 'leiden',
                 resolution: int = 1,
                 aware_spatial: bool = True,
            aware_multi: bool = True,):
        ##spatial type
        if sp_type=='leiden' and 'leiden' not in self.adata_sp.obs.columns:
            sc.pp.neighbors(self.adata_sp,n_neighbors=15,use_rep=self.use_rep_sp)
            sc.tl.leiden(self.adata_sp,resolution=resolution)
            self.adata_sp.obs['type']=self.adata_sp.obs[sp_type].tolist()
        elif sp_type=='louvain' and 'louvain' not in self.adata_sp.obs.columns:
            sc.pp.neighbors(self.adata_sp,n_neighbors=15,use_rep=self.use_rep_sp)
            sc.tl.louvain(self.adata_sp)
            self.adata_sp.obs['type']=self.adata_sp.obs[sp_type].tolist()
        elif sp_type is None:
            aware_spatial=False
        else:
            self.adata_sp.obs['type']=self.adata_sp.obs[sp_type].tolist()

        ##single cell type
        if sc_type is None:
            aware_multi=False
        else:
            self.adata_sc.obs['type']=self.adata_sc.obs[sc_type].tolist()
        self.spatrio_decon=map_info
        return map_info

    def save_map(self,path):
        self.spatrio_decon.to_csv(path)
    
    def loc_prob(self,spot_cell_prob,
                 sc_type: str = 'celltype',
                 sc_prop: float = 0.5,
                 n_cpu=6,):

        map_info=self.spatrio_decon.copy()
        from tqdm import tqdm
        df=spot_cell_prob

        # 创建字典
        first_max_dict = {}
        first_max_value_dict = {}
        
        # 找到每个索引得分第一高的列
        print('...finding spot prop')
        for idx in tqdm(df.index):
            series = df.loc[idx]
            sorted_series = series.sort_values(ascending=False)
            third_max_value = sorted_series.iloc[0]  # 第三大的值
            third_max_column = list(df)[list(series).index(third_max_value)]  # 对应的列名
            first_max_dict[idx] = third_max_column
            first_max_value_dict[idx]=third_max_value

        map_info['spot_prop1']=map_info['spot'].map(first_max_dict)
        map_info['spot_prop_value1']=map_info['spot'].map(first_max_value_dict)

        # 创建字典
        second_max_dict = {}
        second_max_value_dict = {}
        
        # 找到每个索引得分第二高的列
        print('...finding spot prop')
        for idx in tqdm(df.index):
            series = df.loc[idx]
            sorted_series = series.sort_values(ascending=False)
            third_max_value = sorted_series.iloc[1]  # 第三大的值
            third_max_column = list(df)[list(series).index(third_max_value)]  # 对应的列名
            second_max_dict[idx] = third_max_column
            second_max_value_dict[idx]=third_max_value
        map_info['spot_prop2']=map_info['spot'].map(second_max_dict)
        map_info['spot_prop_value2']=map_info['spot'].map(second_max_value_dict)

        # 创建字典
        third_max_dict = {}
        third_max_value_dict = {}

        # 找到每个索引得分第三高的列
        print('...finding spot prop')
        for idx in tqdm(df.index):
            series = df.loc[idx]
            sorted_series = series.sort_values(ascending=False)
            third_max_value = sorted_series.iloc[2]  # 第三大的值
            third_max_column = list(df)[list(series).index(third_max_value)]  # 对应的列名
            third_max_dict[idx] = third_max_column
            third_max_value_dict[idx]=third_max_value

     
        map_info['spot_prop3']=map_info['spot'].map(third_max_dict)
        map_info['spot_prop_value3']=map_info['spot'].map(third_max_value_dict)

        print('...adding spot prop to map info')

        map_info['cell_type']=map_info['cell'].map(dict(zip(
            self.adata_sc.obs.index.tolist(),
            self.adata_sc.obs[sc_type].tolist()   
        )))
        print('...adding cell type to map info')

        cond1 = (map_info['cell_type'] == map_info['spot_prop1']) & (map_info['spot_prop_value1'] > sc_prop)
        print('... finish spot type 1 filter')
        cond2 = (map_info['cell_type'] == map_info['spot_prop2']) & (map_info['spot_prop_value2'] > sc_prop)
        print('... finish spot type 2 filter')
        cond3 = (map_info['cell_type'] == map_info['spot_prop3']) & (map_info['spot_prop_value3'] > sc_prop)
        print('... finish spot type 3 filter')
        mask = cond1 | cond2 | cond3

        print('...filtering map info')
        # 应用过滤条件 ----------------------------------------------------------
        map_info.loc[~mask, 'value'] = -1
        map_info=map_info[map_info['value']!=-1]
        self.spatrio_decon = map_info[['spot', 'cell', 'value']]



    def loc_assign(self,**kwargs):
        """
        Assign spatial coordinates to single cell data.
        """
        spatrio_map = assign_coord(adata1 = self.adata_sp,adata2 = self.adata_sc,
                                   out_data = self.spatrio_decon,**kwargs)
        #self.adata_sp.obs=self.adata_sp.obs.join(spatrio_map.set_index('cell'))

        self.spatrio_map=spatrio_map

        loc1=pd.DataFrame(spatrio_map[['Cell_xcoord','Cell_ycoord','spot','spot_type',
                                       'value']].values,
                          columns=['Cell_xcoord','Cell_ycoord','spot','spot_type','value'],
                          index=spatrio_map['cell'].tolist())
        print('...assigning spatial coordinates to single cell data')
        adata_sc_copy=self.adata_sc.copy()
        adata_sc_copy=adata_sc_copy[loc1.index.tolist()]
        adata_sc_copy.obs['Cell_xcoord']=loc1.loc[adata_sc_copy.obs.index,'Cell_xcoord'].tolist()
        adata_sc_copy.obs['Cell_ycoord']=loc1.loc[adata_sc_copy.obs.index,'Cell_ycoord'].tolist()
        adata_sc_copy.obs['spot_type']=loc1.loc[adata_sc_copy.obs.index,'spot_type'].tolist()
        adata_sc_copy.obs['spot']=loc1.loc[adata_sc_copy.obs.index,'spot'].tolist()
        adata_sc_copy.obs['spot_value']=loc1.loc[adata_sc_copy.obs.index,'value'].tolist()

        print('...adding spatial coordinates to single cell data')
        adata_sc_copy.obsm['spatial']=loc1.loc[adata_sc_copy.obs.index,
                                               ['Cell_xcoord','Cell_ycoord']].values
        adata_sc_copy.uns['spatial']=self.adata_sp.uns['spatial'].copy()
        adata_sc_copy.obsm['spatial']=adata_sc_copy.obsm['spatial'].astype(float)
        del adata_sc_copy.obsm['reduction']
        
        return adata_sc_copy


r"""
Data preprocess and build graph
"""
from typing import Optional, Union, List

import pandas as pd
import numpy as np
import scanpy as sc
import torch
from torch_geometric.nn import knn_graph, radius_graph
from torch_geometric.utils import to_undirected
from anndata import AnnData


def Cal_Spatial_Net(adata:AnnData,
                    rad_cutoff:Optional[Union[None,int]]=None,
                    k_cutoff:Optional[Union[None,int]]=None, 
                    model:Optional[str]='Radius',
                    return_data:Optional[bool]=False,
                    verbose:Optional[bool]=True
    ) -> None:
    r"""
    Construct the spatial neighbor networks.

    Parameters
    ----------
    adata
        AnnData object of scanpy package.
    rad_cutoff
        radius cutoff when model='Radius'
    k_cutoff
        The number of nearest neighbors when model='KNN'
    model
        The network construction model. When model=='Radius', the spot is connected to spots whose distance is less than rad_cutoff. 
        When model=='KNN', the spot is connected to its first k_cutoff nearest neighbors.
    
    Returns
    -------
    The spatial networks are saved in adata.uns['Spatial_Net']
    """

    assert(model in ['Radius', 'KNN'])
    if verbose:
        print('Calculating spatial neighbor graph ...')

    if model == 'KNN':
        edge_index = knn_graph(x=torch.tensor(adata.obsm['spatial']), flow='target_to_source',
                                k=k_cutoff, loop=True, num_workers=8)
        edge_index = to_undirected(edge_index, num_nodes=adata.shape[0]) # ensure the graph is undirected
    elif model == 'Radius':
        edge_index = radius_graph(x=torch.tensor(adata.obsm['spatial']), flow='target_to_source',
                                    r=rad_cutoff, loop=True, num_workers=8) 

    graph_df = pd.DataFrame(edge_index.numpy().T, columns=['Cell1', 'Cell2'])
    id_cell_trans = dict(zip(range(adata.n_obs), adata.obs_names))
    graph_df['Cell1'] = graph_df['Cell1'].map(id_cell_trans)
    graph_df['Cell2'] = graph_df['Cell2'].map(id_cell_trans)
    adata.uns['Spatial_Net'] = graph_df
    
    if verbose:
        print(f'The graph contains {graph_df.shape[0]} edges, {adata.n_obs} cells.')
        print(f'{graph_df.shape[0]/adata.n_obs} neighbors per cell on average.')

    if return_data:
        return adata


def scanpy_workflow(adata:AnnData,
                    filter_cell:Optional[bool]=False,
                    min_gene:Optional[int]=200,
                    min_cell:Optional[int]=30,
                    call_hvg:Optional[bool]=True,
                    n_top_genes:Optional[Union[int, List]]=2500,
                    batch_key:Optional[str]=None,
                    n_comps:Optional[int]=50,
                    viz:Optional[bool]=False,
                    resolution:Optional[float]=0.8
    ) -> AnnData:
    r"""
    Scanpy workflow using Seurat HVG
    
    Parameters
    ----------
    adata
        adata
    filter_cell
        whether to filter cells and genes
    min_gene
        min number of genes per cell
    min_cell
        min number of cells per gene
    call_hvg
        whether to call highly variable genes (only support seurat_v3 method)
    n_top_genes
        n top genes or gene list
    n_comps
        n PCA components
    viz
        whether to run visualize steps
    resolution
        resolution for leiden clustering (used when viz=True)
        
    Return
    ----------
    anndata object
    """
    if 'counts' not in adata.layers.keys():
        adata.layers["counts"] = adata.X.copy()
        
    if filter_cell:
        sc.pp.filter_cells(adata, min_genes=min_gene)
        sc.pp.filter_genes(adata, min_cells=min_cell)

    if call_hvg:
        if isinstance(n_top_genes, int):
            if adata.n_vars > n_top_genes:
                sc.pp.highly_variable_genes(adata, n_top_genes=n_top_genes, flavor="seurat_v3", batch_key=batch_key)
            else:
                adata.var['highly_variable'] = True
                print("All genes are highly variable.")
        elif isinstance(n_top_genes, list):
            adata.var['highly_variable'] = False
            n_top_genes = list(set(adata.var.index).intersection(set(n_top_genes)))
            adata.var.loc[n_top_genes, 'highly_variable'] = True
    else:
        print("Skip calling highly variable genes.")
    
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)
    sc.pp.scale(adata, max_value=10)
    if n_comps > 0:
        sc.tl.pca(adata, n_comps=n_comps, svd_solver="arpack")
        
    if viz:
        sc.pp.neighbors(adata, n_neighbors=15, n_pcs=n_comps)
        sc.tl.umap(adata)
        sc.tl.leiden(adata, resolution=resolution)
    return adata


import time
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from .forceatlas2 import ForceAtlas2

def test_multithreading():
    """Test and benchmark the multithreaded implementation of ForceAtlas2"""
    # Create a random graph
    n_nodes = 500
    n_edges = 2000
    G = nx.gnm_random_graph(n_nodes, n_edges)
    
    # Convert to adjacency matrix
    A = nx.to_scipy_sparse_array(G)
    
    # Single-threaded run
    print("Running single-threaded...")
    start_time = time.time()
    forceatlas2 = ForceAtlas2(
        # Behavior
        outboundAttractionDistribution=True,
        edgeWeightInfluence=1.0,
        
        # Performance
        jitterTolerance=1.0,
        barnesHutOptimize=True,
        barnesHutTheta=1.2,
        multiThreaded=False,
        
        # Tuning
        scalingRatio=2.0,
        strongGravityMode=False,
        gravity=1.0,
        
        # Log
        verbose=True
    )
    pos_single = forceatlas2.forceatlas2(A, iterations=50)
    single_time = time.time() - start_time
    print(f"Single-threaded time: {single_time:.2f} seconds")
    
    # Multi-threaded run
    print("Running multi-threaded...")
    start_time = time.time()
    forceatlas2 = ForceAtlas2(
        # Behavior
        outboundAttractionDistribution=True,
        edgeWeightInfluence=1.0,
        
        # Performance
        jitterTolerance=1.0,
        barnesHutOptimize=True,
        barnesHutTheta=1.2,
        multiThreaded=True,
        
        # Tuning
        scalingRatio=2.0,
        strongGravityMode=False,
        gravity=1.0,
        
        # Log
        verbose=True
    )
    pos_multi = forceatlas2.forceatlas2(A, iterations=50)
    multi_time = time.time() - start_time
    print(f"Multi-threaded time: {multi_time:.2f} seconds")
    
    # Calculate speedup
    speedup = single_time / multi_time
    print(f"Speedup: {speedup:.2f}x")
    
    # Compare final positions to verify correctness
    pos_single_arr = np.array(pos_single)
    pos_multi_arr = np.array(pos_multi)
    
    # The layouts might be different due to thread execution order,
    # but should have similar statistical properties
    print("Single-threaded stats:")
    print(f"  Mean position: ({np.mean(pos_single_arr[:, 0]):.3f}, {np.mean(pos_single_arr[:, 1]):.3f})")
    print(f"  Standard deviation: ({np.std(pos_single_arr[:, 0]):.3f}, {np.std(pos_single_arr[:, 1]):.3f})")
    
    print("Multi-threaded stats:")
    print(f"  Mean position: ({np.mean(pos_multi_arr[:, 0]):.3f}, {np.mean(pos_multi_arr[:, 1]):.3f})")
    print(f"  Standard deviation: ({np.std(pos_multi_arr[:, 0]):.3f}, {np.std(pos_multi_arr[:, 1]):.3f})")
    
    # Visualize the layouts
    plt.figure(figsize=(12, 6))
    
    plt.subplot(121)
    plt.title("Single-threaded")
    nx.draw(G, pos=dict(zip(range(n_nodes), pos_single)), node_size=20, edge_color='gray', alpha=0.6)
    
    plt.subplot(122)
    plt.title("Multi-threaded")
    nx.draw(G, pos=dict(zip(range(n_nodes), pos_multi)), node_size=20, edge_color='gray', alpha=0.6)
    
    plt.tight_layout()
    plt.savefig("forceatlas2_comparison.png")
    plt.close()
    
    print("Visualization saved to forceatlas2_comparison.png")

if __name__ == "__main__":
    test_multithreading() 


import numpy as np
import pandas as pd


def compactness(ad, low_dim_embedding="X_pca", SEACells_label="SEACell"):
    """Compute compactness of each metacell.

    Compactness is defined is the average variance of diffusion components across cells that constitute a metcell.

    :param ad: (Anndata) Anndata object
    :param low_dim_embedding: (str) `ad.obsm` field for constructing diffusion components
    :param SEACell_label: (str) `ad.obs` field for computing diffusion component variances

    :return: `pd.DataFrame` with a dataframe of compactness per metacell

    """
    #import palantir
    from ..palantir.utils import run_diffusion_maps, determine_multiscale_space

    components = pd.DataFrame(ad.obsm[low_dim_embedding]).set_index(ad.obs_names)
    dm_res = run_diffusion_maps(components)
    dc = determine_multiscale_space(dm_res, n_eigs=10)

    return pd.DataFrame(
        dc.join(ad.obs[SEACells_label]).groupby(SEACells_label).var().mean(1)
    ).rename(columns={0: "compactness"})


def separation(
    ad, low_dim_embedding="X_pca", nth_nbr=1, cluster=None, SEACells_label="SEACell"
):
    """Compute separation of each metacell.

    Separation is defined is the distance to the nearest neighboring metacell.

    :param ad: (Anndata) Anndata object
    :param low_dim_embedding: (str) `ad.obsm` field for constructing diffusion components
    :param nth_nbr: (int) Which neighbor to use for computing separation
    :param SEACell_label: (str) `ad.obs` field for computing diffusion component variances

    :return: `pd.DataFrame` with a separation of compactness per metacell

    """
    #import palantir
    from ..palantir.utils import run_diffusion_maps, determine_multiscale_space
    components = pd.DataFrame(ad.obsm[low_dim_embedding]).set_index(ad.obs_names)
    dm_res = run_diffusion_maps(components)
    dc = determine_multiscale_space(dm_res, n_eigs=10)

    # Compute DC per metacell
    metacells_dcs = (
        dc.join(ad.obs[SEACells_label], how="inner").groupby(SEACells_label).mean()
    )

    from sklearn.neighbors import NearestNeighbors

    neigh = NearestNeighbors(n_neighbors=nth_nbr)
    nbrs = neigh.fit(metacells_dcs)
    dists, nbrs = nbrs.kneighbors()
    dists = pd.DataFrame(dists).set_index(metacells_dcs.index)
    dists.columns += 1

    nbr_cells = np.array(metacells_dcs.index)[nbrs]

    metacells_nbrs = pd.DataFrame(nbr_cells)
    metacells_nbrs.index = metacells_dcs.index
    metacells_nbrs.columns += 1

    if cluster is not None:
        # Get cluster type of neighbors to ensure they match the metacell cluster
        clusters = ad.obs.groupby(SEACells_label)[cluster].agg(
            lambda x: x.value_counts().index[0]
        )
        nbr_clusters = pd.DataFrame(clusters.values[nbrs]).set_index(clusters.index)
        nbr_clusters.columns = metacells_nbrs.columns
        nbr_clusters = nbr_clusters.join(pd.DataFrame(clusters))

        clusters_match = nbr_clusters.eq(nbr_clusters[cluster], axis=0)
        return pd.DataFrame(dists[nth_nbr][clusters_match[nth_nbr]]).rename(
            columns={1: "separation"}
        )
    else:
        return pd.DataFrame(dists[nth_nbr]).rename(columns={1: "separation"})


def get_density(ad, key, nth_neighbor=150):
    """Compute cell density as 1/ the distance to the 150th (by default) nearest neighbour.

    :param ad: AnnData object
    :param key: (str) key in ad.obsm to use to build diffusion components on.
    :param nth_neighbor:
    :return: pd.DataFrame containing cell ID and density.
    """
    from sklearn.neighbors import NearestNeighbors
    i#mport palantir
    from ..palantir.utils import run_diffusion_maps, determine_multiscale_space

    neigh = NearestNeighbors(n_neighbors=nth_neighbor)

    if "key" in ad.obsm:
        print(f"Using {key} to compute cell density")
        components = pd.DataFrame(ad.obsm["X_pca"]).set_index(ad.obs_names)
    else:
        raise ValueError(f"Key {key} not present in ad.obsm.")

    diffusion_map_results = run_diffusion_maps(components)
    diffusion_components = determine_multiscale_space(
        diffusion_map_results, n_eigs=8
    )

    nbrs = neigh.fit(diffusion_components)
    cell_density = (
        pd.DataFrame(nbrs.kneighbors()[0][:, nth_neighbor - 1])
        .set_index(ad.obs_names)
        .rename(columns={0: "density"})
    )
    density = 1 / cell_density

    return density


def celltype_frac(x, col_name):
    """TODO."""
    val_counts = x[col_name].value_counts()
    return val_counts.values[0] / val_counts.values.sum()


def compute_celltype_purity(ad, col_name):
    """Compute the purity (prevalence of most abundant value) of the specified col_name from ad.obs within each metacell.

    @param: ad - AnnData object with SEACell assignment and col_name in ad.obs dataframe
    @param: col_name - (str) column name within ad.obs representing celltype groupings for each cell.
    """
    celltype_fraction = ad.obs.groupby("SEACell").apply(
        lambda x: celltype_frac(x, col_name)
    )
    celltype = ad.obs.groupby("SEACell").apply(
        lambda x: x[col_name].value_counts().index[0]
    )

    return pd.concat([celltype, celltype_fraction], axis=1).rename(
        columns={0: col_name, 1: f"{col_name}_purity"}
    )



import numpy as np

from math import ceil
from sklearn.cluster import KMeans
from copy import deepcopy

# All of this functionaliy has been taken from https://github.com/willtownes/nsf-paper/blob/main/utils/preprocess.py
# and https://github.com/willtownes/nsf-paper/blob/main/utils/postprocess.py
def rescale_spatial_coords(X,box_side=4):
  """
  X is an NxD matrix of spatial coordinates
  Returns a rescaled version of X such that aspect ratio is preserved
  But data are centered at zero and area of equivalent bounding box set to
  box_side^D
  Goal is to rescale X to be similar to a N(0,1) distribution in all axes
  box_side=4 makes the data fit in range (-2,2)
  """
  xmin = X.min(axis=0)
  X -= xmin
  x_gmean = np.exp(np.mean(np.log(X.max(axis=0))))
  X *= box_side/x_gmean
  return X - X.mean(axis=0)

def scanpy_sizefactors(Y):
  sz = Y.sum(axis=1,keepdims=True)
  return sz/np.median(sz)

def anndata_to_train_val(ad, layer=None, nfeat=None, train_frac=0.95,
                         sz="constant", dtp="float32", flip_yaxis=True):
  """
  Convert anndata object ad to a training data dictionary
  and a validation data dictionary
  Requirements:
  * rows of ad are pre-shuffled to ensure random split of train/test
  * spatial coordinates in ad.obsm['spatial']
  * features (cols) of ad sorted in decreasing importance (eg with deviance)
  """
  from contextlib import suppress
  if nfeat is not None: ad = ad[:,:nfeat]
  N = ad.shape[0]
  Ntr = round(train_frac*N)
  X = ad.obsm["spatial"].copy().astype(dtp)
  if flip_yaxis: X[:,1] = -X[:,1]
  X = rescale_spatial_coords(X)
  if layer is None: Y = ad.X
  else: Y = ad.layers[layer]
  with suppress(AttributeError):
    Y = Y.toarray() #in case Y is a sparse matrix
  Y = Y.astype(dtp)
  Dtr = {"X":X[:Ntr,:], "Y":Y[:Ntr,:]}
  Dval = {"X":X[Ntr:,:], "Y":Y[Ntr:,:]}
  if sz=="constant":
    Dtr["sz"] = np.ones((Ntr,1),dtype=dtp)
    Dval["sz"] = np.ones((N-Ntr,1),dtype=dtp)
  elif sz=="mean":
    Dtr["sz"] = Dtr["Y"].mean(axis=1,keepdims=True)
    Dval["sz"] = Dval["Y"].mean(axis=1,keepdims=True)
  elif sz=="scanpy":
    Dtr["sz"] = scanpy_sizefactors(Dtr["Y"])
    Dval["sz"] = scanpy_sizefactors(Dval["Y"])
  else:
    raise ValueError("unrecognized size factors 'sz'")
  Dtr["idx"] = np.arange(Ntr)
  if Ntr>=N: Dval = None #avoid returning an empty array
  return Dtr,Dval

def minibatch_size_adjust(num_obs,batch_size):
  """
  Calculate adjusted minibatch size that divides
  num_obs as evenly as possible
  num_obs : number of observations in full data
  batch_size : maximum size of a minibatch
  """
  
  nbatch = ceil(num_obs/float(batch_size))
  return int(ceil(num_obs/nbatch))

def prepare_datasets_tf(Dtrain,Dval=None,shuffle=False,batch_size=None):
  """
  Dtrain and Dval are dicts containing numpy np.arrays of data.
  Dtrain must contain the key "Y"
  Returns a from_tensor_slices conversion of Dtrain and a dict of tensors for Dval
  """
  
  from tensorflow import constant
  from tensorflow.data import Dataset
  Ntr = Dtrain["Y"].shape[0]
  if batch_size is None:
    #ie one batch containing all observations by default
    batch_size = Ntr
  else:
    batch_size = minibatch_size_adjust(Ntr,batch_size)
  Dtrain = Dataset.from_tensor_slices(Dtrain)
  if shuffle:
    Dtrain = Dtrain.shuffle(Ntr)
  Dtrain = Dtrain.batch(batch_size)
  if Dval is not None:
    Dval = {i:constant(Dval[i]) for i in Dval}
  return Dtrain, Ntr, Dval

def kmeans_inducing_pts(X,M):
  M = int(M)
  Z = np.unique(X, axis=0)
  unique_locs = Z.shape[0]
  if M<unique_locs:
    Z=KMeans(n_clusters=M).fit(X).cluster_centers_
  return Z

def t2np(X):
  return X.numpy().mean(axis=0)

def normalize_cols(W):
  """
  Rescale the columns of a matrix to sum to one
  """
  wsum = W.sum(axis=0)
  return W/wsum, wsum

def normalize_rows(W):
  """
  Rescale the rows of a matrix to sum to one
  """
  wsum = W.sum(axis=1)
  return (W.T/wsum).T, wsum

def interpret_nonneg(factors,loadings,lda_mode=False,sort=True):
  """
  Rescale factors and loadings from a nonnegative factorization
  to improve interpretability. Two possible rescalings:

  1. Soft clustering of observations (lda_mode=True):
  Rows of factor matrix sum to one, cols of loadings matrix sum to one
  Returns a dict with keys: "factors", "loadings", and "factor_sums"
  factor_sums is the "n" in the multinomial
  (ie the sum of the counts per observations)

  2. Soft clustering of features (lda_mode=False):
  Rows of loadings matrix sum to one, cols of factors matrix sum to one
  Returns a dict with keys: "factors", "loadings", and "feature_sums"
  feature_sums is similar to an intercept term for each feature
  """
  if lda_mode:
    W,eF,eFsum = rescale_as_lda(factors,loadings,sort=sort)
    return {"factors":eF,"loadings":W,"totals":eFsum}
  else: #spatialDE mode
    eF,W,Wsum = rescale_as_lda(loadings,factors,sort=sort)
    return {"factors":eF,"loadings":W,"totals":Wsum}

def interpret_nsf(fit,X,S=10,**kwargs):
  """
  fit: object of type SF with non-negative factors
  X: spatial coordinates to predict on
  returns: interpretable loadings W, factors eF, and total counts vector
  """
  Fhat = t2np(fit.sample_latent_GP_funcs(X,S=S,chol=False)).T #NxL
  return interpret_nonneg(np.exp(Fhat),fit.W.numpy(),**kwargs)

def rescale_as_lda(factors,loadings,sort=True):
  """
  Rescale nonnegative factors and loadings matrices to be
  comparable to LDA:
  Rows of factor matrix sum to one, cols of loadings matrix sum to one
  """
  W = deepcopy(loadings)
  eF = deepcopy(factors)
  W,wsum = normalize_cols(W)
  eF,eFsum = normalize_rows(eF*wsum)
  if sort:
    o = np.argsort(-eF.sum(axis=0))
    return W[:,o],eF[:,o],eFsum
  else:
    return W,eF,eFsum


import numpy as np
import pandas as pd
import scanpy as sc
import statsmodels.api as sm
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import matplotlib

def volcano(result,pval_name='qvalue',fc_name='log2FC',pval_max=None,FC_max=None,
            figsize:tuple=(4,4),title:str='',titlefont:dict={'weight':'normal','size':14,},
                     up_color:str='#e25d5d',down_color:str='#7388c1',normal_color:str='#d7d7d7',
                     up_fontcolor:str='#e25d5d',down_fontcolor:str='#7388c1',normal_fontcolor:str='#d7d7d7',
                     legend_bbox:tuple=(0.8, -0.2),legend_ncol:int=2,legend_fontsize:int=12,
                     plot_genes:list=None,plot_genes_num:int=10,plot_genes_fontsize:int=10,
                     ticks_fontsize:int=12,pval_threshold:float=0.05,fc_max:float=1.5,fc_min:float=-1.5,
                     ax = None,):
    result=result.copy()
    result['-log(qvalue)']=-np.log10(result[pval_name])
    result['log2FC']= result[fc_name].copy()
    if pval_max!=None:
        result.loc[result['-log(qvalue)']>pval_max,'-log(qvalue)']=pval_max
    if FC_max!=None:
        result.loc[result['log2FC']>FC_max,'log2FC']=FC_max
        result.loc[result['log2FC']<-FC_max,'log2FC']=0-FC_max
    
    if ax==None:
        fig, ax = plt.subplots(figsize=figsize)

    ax.scatter(x=result[result['sig']=='normal']['log2FC'],
            y=result[result['sig']=='normal']['-log(qvalue)'],
            color=normal_color,#颜色
            alpha=.5,#透明度
            )
    #接着绘制上调基因
    ax.scatter(x=result[result['sig']=='up']['log2FC'],
            y=result[result['sig']=='up']['-log(qvalue)'],
            color=up_color,#选择色卡第15个颜色
            alpha=.5,#透明度
            )
    #绘制下调基因
    ax.scatter(x=result[result['sig']=='down']['log2FC'],
            y=result[result['sig']=='down']['-log(qvalue)'],
            color=down_color,#颜色
            alpha=.5,#透明度
            )

    ax.plot([result['log2FC'].min(),result['log2FC'].max()],#辅助线的x值起点与终点
            [-np.log10(pval_threshold),-np.log10(pval_threshold)],#辅助线的y值起点与终点
            linewidth=2,#辅助线的宽度
            linestyle="--",#辅助线类型：虚线
            color='black'#辅助线的颜色
    )
    ax.plot([fc_max,fc_max],
            [result['-log(qvalue)'].min(),result['-log(qvalue)'].max()],
            linewidth=2, 
            linestyle="--",
            color='black')
    ax.plot([fc_min,fc_min],
            [result['-log(qvalue)'].min(),result['-log(qvalue)'].max()],
            linewidth=2, 
            linestyle="--",
            color='black')
    #设置横标签与纵标签
    ax.set_ylabel(r'$-log_{10}(qvalue)$',titlefont)                                    
    ax.set_xlabel(r'$log_{2}FC$',titlefont)
    #设置标题
    ax.set_title(title,titlefont)

    #绘制图注
    #legend标签列表，上面的color即是颜色列表
    labels = ['up:{0}'.format(len(result[result['sig']=='up'])),
            'down:{0}'.format(len(result[result['sig']=='down']))]  
    #用label和color列表生成mpatches.Patch对象，它将作为句柄来生成legend
    color = [up_color,down_color]
    patches = [mpatches.Patch(color=color[i], label="{:s}".format(labels[i]) ) for i in range(len(color))] 

    ax.legend(handles=patches,
        bbox_to_anchor=legend_bbox, 
        ncol=legend_ncol,
        fontsize=legend_fontsize)

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)

    ax.set_xticks([round(i,2) for i in ax.get_xticks()[1:-1]],#获取x坐标轴内容
        [round(i,2) for i in ax.get_xticks()[1:-1]],#更新x坐标轴内容
        fontsize=ticks_fontsize,
        fontweight='normal'
        )

    plt.grid(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)
    ax.spines['left'].set_position(('outward', 10))
    ax.spines['bottom'].set_position(('outward', 10))


    from adjustText import adjust_text
    import adjustText

    if plot_genes is not None:
        hub_gene=plot_genes
    elif (plot_genes is None) and (plot_genes_num is None):
        return ax
    else:
        up_result=result.loc[result['sig']=='up']
        down_result=result.loc[result['sig']=='down']
        hub_gene=up_result.sort_values(pval_name).index[:plot_genes_num//2].tolist()+down_result.sort_values(pval_name).index[:plot_genes_num//2].tolist()

    color_dict={
    'up':up_fontcolor,
        'down':down_fontcolor,
        'normal':normal_fontcolor
    }

    texts=[ax.text(result.loc[i,'log2FC'], 
        result.loc[i,'-log(qvalue)'],
        i,
        fontdict={'size':plot_genes_fontsize,'weight':'bold','color':color_dict[result.loc[i,'sig']]}
        ) for i in hub_gene]

    if adjustText.__version__<='0.8':
        adjust_text(texts,only_move={'text': 'xy'},arrowprops=dict(arrowstyle='->', color='red'),)
    else:
        adjust_text(texts,only_move={"text": "xy", "static": "xy", "explode": "xy", "pull": "xy"},
                    arrowprops=dict(arrowstyle='->', color='red'))

    
    return ax

def venn(sets={}, out='./', palette='bgrc',
             ax=False, ext='png', dpi=300, fontsize=3.5,
             bbox_to_anchor=(.5, .99),nc=2,cs=4):
    
    from ..utils import venny4py
    venny4py(sets=sets,out=out,ce=palette,asax=ax,ext=ext,
             dpi=dpi,size=fontsize,bbox_to_anchor=bbox_to_anchor,
             nc=nc,cs=cs,
             )
    return ax

def boxplot(data,hue,x_value,y_value,width=0.6,title='',
                 figsize=(6,3),palette=None,fontsize=10,
                 legend_bbox=(1, 0.55),legend_ncol=1,):
    """
    Plots a boxplot with jittered points.

    Parameters
    ----------
    - data : `pandas.DataFrame`
        Dataframe containing the data to plot.
    - hue : `str`
        Column name of the dataframe containing the hue data.
    - x_value : `str`
        Column name of the dataframe containing the x-axis data.
    - y_value : `str`
        Column name of the dataframe containing the y-axis data.
    - width : `float`, optional
        Width of the boxplot, by default 0.6.
    - title : `str`, optional
        Title of the plot, by default ''.
    - figsize : `tuple`, optional
        Size of the figure, by default (6,3).
    - palette : `list`, optional
        List of colors to use for the plot, by default None.
    - fontsize : `int`, optional
        Font size of the plot, by default 10.
    - legend_bbox : `tuple`, optional
        Bounding box of the legend, by default (1, 0.55).
    - legend_ncol : `int`, optional
        Number of columns in the legend, by default 1.

    Returns
    -------
    - fig: `matplotlib.figure.Figure`
        Figure object.
    - ax: `matplotlib.axes._subplots.AxesSubplot`
        Axes object.
    """

    #获取需要分割的数据
    hue=hue
    hue_datas=list(set(data[hue]))

    #获取箱线图的横坐标
    x=x_value
    ticks=list(set(data[x]))

    #在这个数据中，我们有6个不同的癌症，每个癌症都有2个基因（2个箱子）
    #所以我们需要得到每一个基因的6个箱线图位置，6个散点图的抖动
    plot_data1={}#字典里的每一个元素就是每一个基因的所有值
    plot_data_random1={}#字典里的每一个元素就是每一个基因的随机20个值
    plot_data_xs1={}#字典里的每一个元素就是每一个基因的20个抖动值


    #箱子的参数
    #width=0.6
    y=y_value
    from ..utils import ticks_range
    import random
    for hue_data,num in zip(hue_datas,ticks_range(len(hue_datas),width)):
        data_a=[]
        data_a_random=[]
        data_a_xs=[]
        for i,k in zip(ticks,range(len(ticks))):
            test_data=data.loc[((data[x]==i)&(data[hue]==hue_data)),y].tolist()
            data_a.append(test_data)
            if len(test_data)<20:
                data_size=len(test_data)
            else:
                data_size=20
            random_data=random.sample(test_data,data_size)
            data_a_random.append(random_data)
            data_a_xs.append(np.random.normal(k*len(hue_datas)+num, 0.04, len(random_data)))
        #data_a=np.array(data_a)
        data_a_random=np.array(data_a_random)
        plot_data1[hue_data]=data_a 
        plot_data_random1[hue_data]=data_a_random
        plot_data_xs1[hue_data]=data_a_xs

    fig, ax = plt.subplots(figsize=figsize)
    #色卡
    if palette==None:
        from ._palette import sc_color
        palette=sc_color
    #palette=["#a64d79","#674ea7"]
    #绘制箱线图
    for hue_data,hue_color,num in zip(hue_datas,palette,ticks_range(len(hue_datas),width)):
        b1=ax.boxplot(plot_data1[hue_data], 
                    positions=np.array(range(len(ticks)))*len(hue_datas)+num, 
                    sym='', 
                    widths=width,)
        plt.setp(b1['boxes'], color=hue_color)
        plt.setp(b1['whiskers'], color=hue_color)
        plt.setp(b1['caps'], color=hue_color)
        plt.setp(b1['medians'], color=hue_color)

        clevels = np.linspace(0., 1., len(plot_data_random1[hue_data]))
        for x, val, clevel in zip(plot_data_xs1[hue_data], plot_data_random1[hue_data], clevels):
            plt.scatter(x, val,c=hue_color,alpha=0.4)

    #坐标轴字体
    #fontsize=10
    #修改横坐标
    ax.set_xticks(range(0, len(ticks) * len(hue_datas), len(hue_datas)), ticks,fontsize=fontsize)
    #修改纵坐标
    yticks=ax.get_yticks()
    ax.set_yticks(yticks[yticks>=0],yticks[yticks>=0],fontsize=fontsize)

    labels = hue_datas  #legend标签列表，上面的color即是颜色列表
    color = palette
    #用label和color列表生成mpatches.Patch对象，它将作为句柄来生成legend
    patches = [ mpatches.Patch(color=color[i], label="{:s}".format(labels[i]) ) for i in range(len(hue_datas)) ] 
    ax.legend(handles=patches,bbox_to_anchor=legend_bbox, ncol=legend_ncol,fontsize=fontsize)

    #设置标题
    ax.set_title(title,fontsize=fontsize+1)
    #设置spines可视化情况
    plt.grid(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)
    ax.spines['left'].set_position(('outward', 10))
    ax.spines['bottom'].set_position(('outward', 10))
    
    return fig,ax


import numpy as np
from scipy import sparse
from sys import getsizeof

from scipy.stats import spearmanr, pearsonr
from sklearn.ensemble import RandomForestRegressor, GradientBoostingRegressor
from sklearn.preprocessing import normalize
from sklearn.neighbors import NearestNeighbors
import igraph
import networkx as nx

def pairwise_scc(X1, X2):
    X1 = X1.argsort(axis=1).argsort(axis=1)
    X2 = X2.argsort(axis=1).argsort(axis=1)
    X1 = (X1-X1.mean(axis=1, keepdims=True))/X1.std(axis=1, keepdims=True)
    X2 = (X2-X2.mean(axis=1, keepdims=True))/X2.std(axis=1, keepdims=True)
    sccmat = np.empty([X1.shape[0], X2.shape[0]], float)
    for i in range(X1.shape[0]):
        for j in range(X2.shape[0]):
            c = np.dot( X1[i,:], X2[j,:]) / float(X1.shape[1])
            sccmat[i,j] = c
    return sccmat

def standardize(X):
    X_mean = X.mean(axis=0)
    X_std = X.std(axis=0)
    X_standard = np.empty_like(X)
    for i in range(X.shape[1]):
        if X_std[i] > 0:
            X_standard[:,i] = (X[:,i] - X_mean[i]) / X_std[i]
        else:
            X_standard[:,i] = 0.0
    return X_standard

def partial_corr(x, y, cov, method="spearman", aggregate=True):
    # Standardize data
    x = standardize(x)
    y = standardize(y)
    cov = standardize(cov)
    if aggregate:
        rs = []
        for i in range(cov.shape[1]):
            beta_x = np.linalg.lstsq(cov[:,i].reshape(-1,1), x, rcond=None)[0]
            beta_y = np.linalg.lstsq(cov[:,i].reshape(-1,1), y, rcond=None)[0]
            res_x = x - cov[:,i].reshape(-1,1) @ beta_x
            res_y = y - cov[:,i].reshape(-1,1) @ beta_y
            if method == "spearman":
                r,p = spearmanr(res_x, res_y)
            elif method == "pearson":
                r,p = pearsonr(res_x, res_y)
            rs.append(r)
        r = np.mean(rs); p = -1
    else:
        beta_x = np.linalg.lstsq(cov, x, rcond=None)[0]
        beta_y = np.linalg.lstsq(cov, y, rcond=None)[0]
        res_x = x - cov @ beta_x
        res_y = y - cov @ beta_y
        if method == "spearman":
            r,p = spearmanr(res_x, res_y)
        elif method == "pearson":
            r,p = pearsonr(res_x, res_y)
    return r, p

def semipartial_corr(x, y, xcov=None, ycov=None, method="spearman"):
    # Standardize data
    x = standardize(x)
    y = standardize(y)
    if not xcov is None:
        xcov = standardize(xcov)
    if not ycov is None:
        ycov = standardize(ycov)
    res_x = np.copy(x); res_y = np.copy(y)
    if not xcov is None:
        beta_x = np.linalg.lstsq(xcov, x, rcond=None)[0]
        res_x = x - xcov @ beta_x
    if not ycov is None:
        beta_y = np.linalg.lstsq(ycov, y, rcond=None)[0]
        res_y = y - ycov @ beta_y
    if method == "spearman":
        r,p = spearmanr(res_x, res_y)
    elif method == "pearson":
        r,p = pearsonr(res_x, res_y)
    return r, p

def treebased_score(x,
    y,
    cov,
    method="rf",
    n_trees=100,
    n_repeat=10,
    max_depth=5,
    max_features="sqrt",
    learning_rate=0.1,
    subsample=1.0
):
    random_seeds = np.random.randint(100000, size=n_repeat)
    X_train = np.concatenate((x.reshape(-1,1), cov), axis=1)
    ranks = []
    for i in range(n_repeat):
        if method == "rf":
            model = RandomForestRegressor(n_estimators=n_trees, 
                max_depth=max_depth,
                max_features=max_features,
                random_state=random_seeds[i],
                n_jobs=-1
            )
        elif method == "gbt":
            model = GradientBoostingRegressor(n_estimators=n_trees,
                max_depth=max_depth,
                max_features=max_features,
                learning_rate=learning_rate,
                subsample=subsample,
                random_state=random_seeds[i]
            )
        model.fit(X_train, y)
        importance = model.feature_importances_
        rank = np.where(np.argsort(-importance)==0)[0][0]
        ranks.append(float(cov.shape[1]-rank)/float(cov.shape[1]))
    return np.mean(ranks)

def treebased_score_multifeature(
    X,
    y,
    cov,
    method='rf',
    n_trees=100,
    n_repeat=10,
    max_depth=5,
    max_features='sqrt',
    learning_rate=0.1,
    subsample=1.0
):
    random_seeds = np.random.randint(100000, size=n_repeat)
    X_train = np.concatenate((X, cov), axis=1)
    ranks = np.empty([X.shape[1], n_repeat], float)
    for i in range(n_repeat):
        if method == 'rf':
            model = RandomForestRegressor(n_estimators=n_trees, 
                max_depth=max_depth,
                max_features=max_features,
                random_state=random_seeds[i],
                n_jobs=-1
            )
        elif method == 'gbt':
            model = GradientBoostingRegressor(n_estimators=n_trees,
                max_depth=max_depth,
                max_features=max_features,
                learning_rate=learning_rate,
                subsample=subsample,
                random_state=random_seeds[i]
            )
        model.fit(X_train, y)
        importance = model.feature_importances_
        sorted_idx = np.argsort(-importance)
        for j in range(X.shape[1]):
            rank = np.where(sorted_idx==j)[0][0]
            ranks[j,i] = float(X_train.shape[1] - rank - 1) / float(X_train.shape[1]-1)
    return np.mean(ranks, axis=1)
        

def vf_diff(vf_1, vf_2):
    norm_vf_1 = np.linalg.norm(vf_1, axis=1)
    norm_vf_2 = np.linalg.norm(vf_2, axis=1)
    dot_vf = (vf_1 * vf_2).sum(axis=1)
    
    s_cosine = dot_vf / (norm_vf_1 * norm_vf_2)
    s_cosine[np.isnan(s_cosine)] = 0.0

    s_coop = dot_vf

    return s_cosine, s_coop

def d_graph_local_jaccard(A1, A2):
    a = float( np.sum( (A1 != 0) * (A2 != 0) ) )
    b = float( np.sum( (A1 != 0) + (A2 != 0) ) )
    if b == 0:
        d = 1
    else:
        d = a / b
    return 1 - d

def d_graph_local_jaccard_weighted(A1, A2):
    a = float( np.sum( np.minimum(A1, A2) ) )
    b = float( np.sum( np.maximum(A1, A2) ) )
    if b == 0:
        d = 1
    else:
        d = a / b
    return 1 - d

# %%
def d_graph_mesoscale_heat(A1, A2, tau, p=2):
    A1_lap = sparse.csgraph.laplacian(A1, normed = True)
    w,v = np.linalg.eigh(A1_lap.todense())
    # tmp_idx = np.argsort(w)
    # w = w[tmp_idx]
    # v = v[:,tmp_idx]
    W = np.diag(w)
    R1 = v.real @ np.exp( - tau * W ) @ v.real.T

    A2_lap = sparse.csgraph.laplacian(A2, normed = True)
    w,v = np.linalg.eigh(A2_lap.todense())
    # tmp_idx = np.argsort(w)
    # w = w[tmp_idx]
    # v = v[:,tmp_idx]
    W = np.diag(w)
    R2 = v.real @ np.exp( - tau * W ) @ v.real.T

    delta = R1 - R2

    N = A1.shape[0]

    d = np.trace(delta.T @ delta) / float(N)

    # d = np.linalg.norm(R1-R2) ** 2 / float(N)
    
    return d

# %%
def d_graph_global_structure(A1, A2, w1=0.45, w2=0.45, w3=0.1):
    # Ref: Schieber, Tiago A., et al. "Quantification of network 
    # structural dissimilarities." Nature communications 8.1 (2017): 1-10.
    A1 = A1.astype(bool).astype(int)
    A2 = A2.astype(bool).astype(int)
    g1 = igraph.Graph(directed=True).Adjacency( (A1>0).tolist() )
    g2 = igraph.Graph(directed=True).Adjacency( (A2>0).tolist() )
    n = g1.vcount()
    m = g2.vcount()
    pg1 = nnd(g1)
    pg2 = nnd(g2)
    pm = np.zeros([max(m,n)], float)
    pm[:n-1] = pg1[:n-1]
    pm[-1] = pg1[n-1]
    pm[:m-1] = pm[:m-1] + pg2[:m-1]
    pm[-1] = pm[-1]+pg2[m-1]
    pm = pm / 2
    first = np.sqrt( max( (entropia(pm)-(entropia(pg1[:n])+entropia(pg2[:m]))/2)/np.log(2), 0) )
    second = np.abs(np.sqrt(pg1[n]) - np.sqrt(pg2[m]))

    pg1 = alpha(A1)
    pg2 = alpha(A2)
    m = max(len(pg1), len(pg2))
    Pg1 = np.zeros([m], float)
    Pg2 = np.zeros([m], float)
    Pg1[m-len(pg1):] = pg1[:]
    Pg2[m-len(pg2):] = pg2[:]
    third = np.sqrt( (entropia((Pg1+Pg2)/2)-(entropia(pg1)+entropia(pg2))/2 )/np.log(2))/2
    A1_compl = np.ones_like(A1) - A1
    for i in range(A1.shape[0]): A1_compl[i,i] = 1
    A2_compl = np.ones_like(A2) - A2
    for i in range(A2.shape[0]): A2_compl[i,i] = 1
    pg1 = alpha(A1_compl)
    pg2 = alpha(A2_compl)
    m = max(len(pg1), len(pg2))
    Pg1 = np.zeros([m], float)
    Pg2 = np.zeros([m], float)
    Pg1[m-len(pg1):] = pg1[:]
    Pg2[m-len(pg2):] = pg2[:]
    third = third + np.sqrt( (entropia((Pg1+Pg2)/2)-(entropia(pg1)+entropia(pg2))/2 )/np.log(2))/2

    return w1*first + w2*second + w3*third


def alpha(A):
    n = A.shape[0]
    g_nx = nx.from_numpy_array(A, create_using=nx.DiGraph)
    g_degree = dict( g_nx.degree() )
    for key in g_degree.keys():
        g_degree[key] = g_degree[key] / (n-1)
    g_nx.remove_edges_from(nx.selfloop_edges(g_nx))
    centrality = nx.katz_centrality(g_nx, beta=g_degree, alpha=1/float(n), normalized=False)
    centrality_vec = []
    for i in range(n):
        centrality_vec.append(centrality[i])
    centrality_vec = np.array(centrality_vec)
    r = np.sort(centrality_vec) / n**2
    rr = np.zeros([n+1],float)
    rr[:n] = r[:]
    rr[-1] = max(0,1-r.sum())
    return rr

def entropia(a):
    a = a[np.where(a>0)]
    return -np.sum(a*np.log(a))

def node_distance(g):
    n = g.vcount()
    a = np.zeros([n,n], float)
    m = np.array( g.shortest_paths(mode="ALL"), float )
    m[np.isinf(m)] = n
    quem = np.unique(m)
    quem = quem[np.where(quem!=0)]
    for mm in quem:
        tmp_mask = (m == mm)
        a[:,int(mm)-1] = tmp_mask.astype(float).sum(axis=1)
    return a / (n-1)

def nnd(g):
    n = g.vcount()
    nd = node_distance(g)
    pdfm = nd.mean(axis=0)
    norm = np.log( max(2, len(np.where(pdfm[:-1]>0)[0])+1) )
    y = max(0, entropia(pdfm)-entropia(nd)/n)/norm
    out = np.zeros([len(pdfm)+1],float)
    out[:n] = pdfm[:]; out[n] = y
    return out

# %%
def spatial_weight(
    X,
    bandwidth = None,
    k = 5,
    function = 'gaussian',
    row_standardize = False,
    zero_diagonal = False
):
    import libpysal
    kw = libpysal.weights.Kernel(X,
        bandwidth = bandwidth, k = k, function = function)
    if row_standardize:
        kw.transform = 'r'
    I = []; J = []; W = []
    for i in range(X.shape[0]):
        nnb = len(kw.neighbors[i])
        for j in range(nnb):
            if zero_diagonal and i == kw.neighbors[i][j]:
                continue
            I.append(i)
            J.append(kw.neighbors[i][j])
            W.append(kw.weights[i][j])
    I = np.array(I, int)
    J = np.array(J, int)
    W = np.array(W, float)
    
    return I, J, W

# %%
def tmp_moranI_vector_global(
    u,
    v,
    I,
    J,
    W
):
    """
    Liu, Yu, Daoqin Tong, and Xi Liu. "Measuring spatial autocorrelation
    of vectors." Geographical Analysis 47.3 (2015): 300-319.
    """
    n = float( len(u) )
    a = np.sum( W * (u[I]*u[J]+v[I]*v[J]) )
    b = np.sum( u**2 + v**2 )
    c = np.sum( W )
    moranI = ( n * a ) / ( b * c )
    return moranI

def tmp_moranI_vector_global_3d(
    u,
    v,
    w,
    I,
    J,
    W
):
    """
    Liu, Yu, Daoqin Tong, and Xi Liu. "Measuring spatial autocorrelation
    of vectors." Geographical Analysis 47.3 (2015): 300-319.
    """
    n = float( len(u) )
    a = np.sum( W * (u[I]*u[J]+v[I]*v[J] + w[I]*w[J]) )
    b = np.sum( u**2 + v**2 + w**2)
    c = np.sum( W )
    moranI = ( n * a ) / ( b * c )
    return moranI

def tmp_moranI_vector_local(
    u,
    v,
    I,
    J,
    W
):
    n = float( len(u) )
    m2 = np.sum( u**2 + v**2 ) / n
    print(m2)
    moranI = np.zeros([len(u)])
    for i in range(len(I)):
        moranI[I[i]] += u[I[i]] * W[i] * u[J[i]] + v[I[i]] * W[i] * v[J[i]]
    moranI = moranI / m2
    return moranI

def moranI_vector_global(
    X,
    V,
    weight_bandwidth = None,
    weight_k = 5,
    weight_function = 'triangular',
    weight_row_standardize = False,
    n_permutations = 999
):
    z_V = V - V.mean(axis=0)
    I, J, W = spatial_weight(X,
        bandwidth = weight_bandwidth, k = weight_k,
        function = weight_function, row_standardize = weight_row_standardize,
        zero_diagonal = True)
    if V.shape[1] == 2: 
        moranI = tmp_moranI_vector_global(z_V[:,0], z_V[:,1], I, J, W)
    elif V.shape[1] == 3:
        moranI = tmp_moranI_vector_global_3d(z_V[:,0], z_V[:,1], z_V[:,2], I, J, W)
    permute_moranI = []
    idx = np.arange(X.shape[0])
    for i in range(n_permutations):
        permute_idx = np.random.permutation(idx)
        if V.shape[1] == 2:
            tmp_I = tmp_moranI_vector_global(z_V[permute_idx,0], z_V[permute_idx,1], I, J, W)
        elif V.shape[1] == 3:
            tmp_I = tmp_moranI_vector_global_3d(z_V[permute_idx,0], z_V[permute_idx,1], z_V[permute_idx,2], I, J, W)
        permute_moranI.append(tmp_I)
    permute_moranI = np.abs( np.array(permute_moranI, float) )
    p_value = ( permute_moranI > np.abs(moranI) ).astype(float).sum() / float(n_permutations)
    
    return moranI, p_value

def moranI_vector_local(
    X,
    V,
    weight_bandwidth = None,
    weight_k = 5,
    weight_function = 'triangular',
    weight_row_standardize = False
):
    z_V = V - V.mean(axis=0)
    I, J, W = spatial_weight(X,
        bandwidth = weight_bandwidth, k = weight_k,
        function = weight_function, row_standardize = weight_row_standardize,
        zero_diagonal = True)
    moranI = tmp_moranI_vector_local(z_V[:,0], z_V[:,1], I, J, W)

    return moranI




# %%

def preprocess_vector_field(
    X,
    V,
    knn_smoothing = -1,
    normalize_vf = 'quantile',
    quantile = 0.99
):
    if knn_smoothing >= 2:
        nbrs = NearestNeighbors(n_neighbors=knn_smoothing, algorithm='ball_tree').fit(X)
        _, idx = nbrs.kneighbors(X)
        V_smoothed = np.zeros_like(V)
        for i in range(knn_smoothing):
            V_smoothed = V_smoothed + V[idx[:,i],:]
        V_smoothed = V_smoothed / float(knn_smoothing)
        V = V_smoothed
    V_norm = np.linalg.norm(V, axis=1)
    if normalize_vf == 'quantile':
        scale = np.quantile(V_norm, quantile)
        if scale > 0:
            V = V / scale
    elif normalize_vf == 'unit_norm':
        V = normalize(V)
    return V

# def d_vf_cosine_similarity(
#     X,
#     V1,
#     V2,
#     knn_smoothing = -1,
#     normalize = "quantile",
#     quantile = 0.99
# ):
#     if knn_smoothing >= 2:
#         nbrs = NearestNeighbors(n_neighbors=knn_smoothing, algorithm='ball_tree').fit(X)
#         _, idx = nbrs.kneighbors(X)
#         V_smoothed = np.zero_like(V)


import numpy as np
import scipy as s
import scipy.stats as stats
from .basic_distributions import Distribution

from ... import config

class UnivariateGaussian(Distribution):
    """
    Class to define univariate Gaussian distributions

    Equations:
    Class for a univariate Gaussian distributed node
    p(x|mu,sigma^2) = 1/sqrt(2*pi*sigma^2) * exp(-0.5*(x-mu)^2/(sigma^2) )
    log p(x|mu,sigma^2) =
    E[x] = mu
    var[x] = sigma^2
    H[x] = 0.5*log(sigma^2) + 0.5*(1+log(2pi))

    """
    def __init__(self, dim, mean, var, E=None, E2=None):
        Distribution.__init__(self, dim)

        # Initialise parameters
        mean = np.ones(dim) * mean
        var = np.ones(dim) * var
        self.params = { 'mean':mean, 'var':var }

        # Initialise expectations
        self.expectations = {}
        if E is None:
            self.updateExpectations()
        else:
            self.expectations['E'] = np.ones(dim)*E

        if E2 is not None:
            self.expectations['E2'] = np.ones(dim)*E2

        # float64 -> float32
        if config.use_float32: self.to_float32()

        # Check that dimensionalities match
        self.CheckDimensionalities()

    def updateExpectations(self):
        # Update first and second moments using current parameters
        E = self.params['mean']
        E2 = E**2 + self.params['var']
        self.expectations = { 'E':E, 'E2':E2 }

    def density(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        # print stats.norm.pdf(x, loc=self.mean, scale=np.sqrt(self.var))
        return np.sum( (1/np.sqrt(2*np.pi*self.params['var'])) * np.exp(-0.5*(x-self.params['mean'])**2/self.params['var']) )

    def loglik(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        # return np.log(stats.norm.pdf(x, loc=self.mean, scale=np.sqrt(self.var)))
        return np.sum( -0.5*np.log(2*np.pi) - 0.5*np.log(self.params['var']) -0.5*(x-self.params['mean'])**2/self.params['var'] )

    def entropy(self):
        return np.sum( 0.5*np.log(self.params['var']) + 0.5*(1+np.log(2*np.pi)) )

    def sample(self):
        return np.random.normal(self.params['mean'], np.sqrt(self.params['var']))




SELECTED_COLOR = "#377EB8"
DESELECTED_COLOR = "#CFD5E2"

# This global variable sets the default behaviour for saving pandas.DataFrames
# in AnnData.obsm and AnnData.varm. When set to True, the data is saved as pandas.DataFrame.
SAVE_AS_DF = True



import os

import numpy as np
import pandas as pd
import scanpy as sc


_ROOT = os.path.abspath(os.path.dirname(__file__))


def get_data(path):
    """Get absolute path relative to package installation for loading static files.

    :param path: (str) relative path to items
    :return: (str) absolute path to items.
    """
    return os.path.join(_ROOT, "data", path)


def get_Rscript(path):
    """Get absolute path for loading Rscripts.

    :param path: (str) relative path to items
    :return: (str) absolute path to items.
    """
    return os.path.join(_ROOT, "Rscripts", path)


def load_data():
    """Get absolute path for loading sample data.

    :return: (anndata.AnnData object) sample dataset.
    """
    p = get_data("sample_data.h5ad")
    return sc.read(p)


def chromVAR_R(outdir):
    """Run chromVAR R scripts.

    Given an output directory containing:
        - peaks.bed
        - sampling_depth.txt
        - peak_names.txt
        - cell_names.txt
        - counts.txt
    Executes chromVAR R script and writes output files to same directory:
        - deviations.csv
        - variability.csv
        - deviationScores.csv.

    Loads and returns
    (1) pd.Dataframe containing deviations and
    (2) pd.Dataframe containing variability
    """
    import subprocess

    result = subprocess.run(
        ["Rscript", get_Rscript("chromVAR.R"), outdir], stdout=subprocess.PIPE
    )
    print("Executing command:", " ".join(["Rscript", "run_chromVAR.R", outdir]))
    from IPython.display import display
    display(result.stdout.decode("utf-8"))

    deviations = pd.read_csv(outdir + "deviations.csv", index_col=[0])
    variability = pd.read_csv(outdir + "variability.csv", index_col=[0])

    return deviations, variability


def run_chromVAR(ad, outdir):
    """Run chromVAR Rscript on anndata object and output deviations to outdir.

    Writes the following files in outdir:
        - peaks.bed
        - sampling_depth.txt
        - peak_names.txt
        - cell_names.txt
        - counts.txt.

    Executes chromVAR_R() - runs chromVAR R script and writes output files to same directory:
        - deviations.csv
        - variability.csv
        - deviationScores.csv

    Loads and returns
    (1) pd.Dataframe containing deviations and
    (2) pd.Dataframe containing variability
    """
    # Create output directory
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    # Convert ad to counts matrix
    atac_df = ad.to_df()

    # Drop zero-count peaks
    atac_df = atac_df.T[atac_df.T.sum(1) > 0]
    ad[:, atac_df.index].var[["seqnames", "start", "end"]].to_csv(
        outdir + "peaks.bed", index=False, sep="\t", header=None
    )
    atac_df.to_csv(outdir + "counts.txt", index=True, sep="\t")
    atac_df.sum(0).to_csv(outdir + "sampling_depth.txt", index=True, sep="\t")

    np.savetxt(outdir + "peak_names.txt", atac_df.index.values, fmt="%s")
    np.savetxt(outdir + "cell_names.txt", atac_df.columns.values, fmt="%s")

    print("Finished writing input files to chromVAR...")

    return chromVAR_R(outdir)


def tanay_metacells():
    """TOOD."""
    raise NotImplementedError


def run_tanay():
    """TODO."""
    raise NotImplementedError



import numpy as np

def filter_genes(counts_mat, gene_labels, umi_threshold=500, exclude_prefix=['MT-', 'RPL', 'RPS']):
    idx_kept=np.where(np.sum(counts_mat,0) > umi_threshold)[0]
    idx_kept=np.array( [i for i in idx_kept if gene_labels[i][:3] not in exclude_prefix] )
    gene_labels_idx=gene_labels[idx_kept]
    return idx_kept, gene_labels_idx


import pandas as pd
import numpy as np
import anndata as ad

def atac_concat_get_index(adata1,adata2):
    import gc
    
    # 数据预处理
    adata1.var.loc[:,'ATAC'] = adata1.var.index
    adata1.var = pd.DataFrame((x.split('_') for x in adata1.var['ATAC']) , index = adata1.var.index , columns = ['chr','chromStart','chromEnd'])
    adata2.var.loc[:,'ATAC'] = adata2.var.index
    adata2.var = pd.DataFrame((x.split('_') for x in adata2.var['ATAC']) , index = adata2.var.index , columns = ['chr','chromStart','chromEnd'])

    # 判断是否出现 不存在某条染色体的情况，如果存在，我们希望染色体较少的一方为adata1（因为后续的分析都是针对adata1为染色体较少的情况进行分析的）
    # 如果出现adata1的染色体比adata2的染色体多，则调换位置
    if len(set(adata1.var.chr))>len(set(adata2.var.chr)):
        adata_chr = adata1
        adata1 = adata2
        adata2 = adata_chr
        Exchange = True #Exchange用于判断调换位置的特殊情况
    else:
        Exchange = False
 

    # 循环函数确定和拆分
    atac_var = adata1.var
    Promoter = adata2.var
    Promoter.loc[:,'chr']=Promoter.loc[:,'chr']+'_'
    Promoter.loc[:,'index_1'] = Promoter.index
    atac_var.loc[:,'chr']=atac_var.loc[:,'chr']+'_'
    atac_var.loc[:,'index_1'] = atac_var.index
    chr	= set(Promoter.chr)

    # 比对函数开始
    ATAC = {}
    PROMOTER = {}
    for x in chr:
        # 染色体拆分
        test = pd.DataFrame(columns=['Length'])
        ATAC[x] = test
        ATAC[x]=atac_var[atac_var.index.str.contains(x)]
        PROMOTER[x] = test
        test = pd.DataFrame(columns=['Length'])
        PROMOTER[x]=Promoter.loc[Promoter['chr'] == x]

        # 输出循环染色体
        print('start %s' % (x))
    
        # 特殊情况处理-判断该条染色体是否为空值,并进行赋值
        if len(ATAC[x].index)==0:
            ATAC[x] = PROMOTER[x]
            ATAC_INDEX = ATAC[x].columns.to_list() + ['overleaf_1','overleaf_2','chromStart_New','chromEnd_New','index_2']
            ATAC[x] = ATAC[x].reindex(columns=ATAC_INDEX, fill_value='retain')
            ATAC[x].loc[:,'index_2'] = ATAC[x].loc[:,'index_1']
            continue
        

        # index赋值
        ATAC_INDEX = ATAC[x].columns.to_list() + ['overleaf_1','overleaf_2','chromStart_New','chromEnd_New','index_2']
        ATAC[x] = ATAC[x].reindex(columns=ATAC_INDEX, fill_value='')
        
        # 函数主循环体
        for i in range(0,len(PROMOTER[x].chromEnd)):
            left = 0
            right = len(ATAC[x].chromEnd)-1
            target_Start = int(PROMOTER[x].chromStart.iloc[i])
            target_End = int(PROMOTER[x].chromEnd.iloc[i])


            while(left<=right):
                # 中位数表示 #
                midIndex = round(int((left+right)/2))
                midValue_Start = int(ATAC[x].chromStart.iloc[midIndex])
                midValue_End = int(ATAC[x].chromEnd.iloc[midIndex])
                # 三次condition判断 #

                if (target_Start > midValue_End) :
                    left = midIndex+1

                elif (target_End < midValue_Start):
                    right = midIndex-1
                else:
                    Length = target_End - target_Start
                    ATAC[x].loc[:,'overleaf_1'].iloc[midIndex] = (target_Start - midValue_End)/Length
                    ATAC[x].loc[:,'overleaf_2'].iloc[midIndex] = (midValue_Start - target_End)/Length
                    list = [midValue_Start,midValue_End,target_Start,target_End]
                    ATAC[x].loc[:,'chromStart_New'].iloc[midIndex] = min(list)
                    ATAC[x].loc[:,'chromEnd_New'].iloc[midIndex] = max(list)
                    ATAC[x].loc[:,'index_2'].iloc[midIndex] = PROMOTER[x].index[i]
                    break
 
    # 拼接dataframe
    ATAC_Promoter = pd.DataFrame()
    for i in chr:
        ATAC_Promoter = pd.concat([ATAC_Promoter,ATAC[i],],axis=0)
        
    # 删除未比对上序列
    ATAC_Promoter = ATAC_Promoter[ATAC_Promoter['overleaf_1']!='']

    # 定义新序列拼接函数
    def indeX(series):
        chr = series["chr"]
        chromStart_New = series["chromStart_New"]
        chromEnd_New = series["chromEnd_New"]
        indeX = chr + str(chromStart_New) +'_' +str(chromEnd_New)
        return indeX

    # 拼接新序列
    ATAC_Promoter.loc[:,'index_new'] = ''
    ATAC_Promoter.loc[:,'index_new'] = ATAC_Promoter.apply(indeX,axis=1)

    # 提取新序列
    pair_index = pd.DataFrame()
    pair_index.loc[:,'overleaf_1'] = ATAC_Promoter.loc[:,'overleaf_1']
    pair_index.loc[:,'overleaf_2'] = ATAC_Promoter.loc[:,'overleaf_2']
    pair_index.loc[:,'index_new'] = ATAC_Promoter.loc[:,'index_new']
    # 若出现adata1调换adata2的特殊情况，则在这里把adata1和adata2再调换回来
    if Exchange==True:
        pair_index.loc[:,'index_1'] = ATAC_Promoter.loc[:,'index_2']
        pair_index.loc[:,'index_2'] = ATAC_Promoter.loc[:,'index_1']
    else:
        pair_index.loc[:,'index_1'] = ATAC_Promoter.loc[:,'index_1']
        pair_index.loc[:,'index_2'] = ATAC_Promoter.loc[:,'index_2']
    # 特殊情况处理-空值染色体
    pair_index.loc[pair_index['overleaf_1']=='retain','index_new'] = pair_index.loc[pair_index['overleaf_1']=='retain','index_1']
    gc.collect()
    return pair_index


def atac_concat_inner(adata1,adata2,pair_index):
    import gc
    
    # 设置基本参数
    adata1.var.loc[:,'ATAC'] = adata1.var.index
    df1 = pd.DataFrame((x.split('_') for x in adata1.var['ATAC']) , index = adata1.var.index , columns = ['chr','chromStart','chromEnd'])
    adata1.var = df1
    adata2.var.loc[:,'ATAC'] = adata2.var.index
    df2 = pd.DataFrame((x.split('_') for x in adata2.var['ATAC']) , index = adata2.var.index , columns = ['chr','chromStart','chromEnd'])
    adata2.var = df2

    # 判断是否存在需要调换位置的情况
    if len(set(adata1.var.chr))>len(set(adata2.var.chr)): 
        Exchange = True #Exchange用于判断调换位置的特殊情况
    else:
        Exchange = False
    
    # 通过overleaf_1的赋值来判断是否存在缺失染色体的情况
    if Exchange==False:
        atac1_new = adata1[:,pair_index[pair_index['overleaf_1']!='retain'].index_1]
        atac1_new.var = pair_index[pair_index['overleaf_1']!='retain']
        atac1_new.var.index = atac1_new.var.index_new
        atac2_new = adata2[:,pair_index.index_2]
        atac2_new.var = pair_index
        atac2_new.var.index = atac2_new.var.index_new

    if Exchange==True:
        atac1_new = adata1[:,pair_index.index_1]
        atac1_new.var = pair_index
        atac1_new.var.index = atac1_new.var.index_new
        atac2_new = adata2[:,pair_index[pair_index['overleaf_1']!='retain'].index_2]
        atac2_new.var = pair_index[pair_index['overleaf_1']!='retain']
        atac2_new.var.index = atac2_new.var.index_new
    
    # 配对
    adata_pair = ad.concat([atac1_new,atac2_new],axis=0,join='outer',fill_value=0)
    adata_pair.obs_names_make_unique()
    adata_pair
    gc.collect()
    return adata_pair

def atac_concat_outer(adata1,adata2,pair_index):
    import gc
    
        # 提取关键参数
    adata1.var.loc[:,'ATAC'] = adata1.var.index
    adata1.var = pd.DataFrame((x.split('_') for x in adata1.var['ATAC']) , index = adata1.var.index , columns = ['chr','chromStart','chromEnd'])
    adata2.var.loc[:,'ATAC'] = adata2.var.index
    adata2.var = pd.DataFrame((x.split('_') for x in adata2.var['ATAC']) , index = adata2.var.index , columns = ['chr','chromStart','chromEnd'])

        # 判断是否存在需要调换位置的情况
    if len(set(adata1.var.chr))>len(set(adata2.var.chr)): 
        Exchange = True #Exchange用于判断调换位置的特殊情况
    else:
        Exchange = False
    
        # 通过overleaf_1的赋值来判断是否存在缺失染色体的情况
    if Exchange==False:
        atac1_new = adata1[:,pair_index[pair_index['overleaf_1']!='retain'].index_1]
        atac1_new.var = pair_index[pair_index['overleaf_1']!='retain']
        atac1_new.var.index = atac1_new.var.index_new
        atac2_new = adata2[:,pair_index.index_2]
        atac2_new.var = pair_index
        atac2_new.var.index = atac2_new.var.index_new

    if Exchange==True:
        atac1_new = adata1[:,pair_index.index_1]
        atac1_new.var = pair_index
        atac1_new.var.index = atac1_new.var.index_new
        atac2_new = adata2[:,pair_index[pair_index['overleaf_1']!='retain'].index_2]
        atac2_new.var = pair_index[pair_index['overleaf_1']!='retain']
        atac2_new.var.index = atac2_new.var.index_new

    # 计算前置数据
    index_list = atac1_new.var
    index_list.index = index_list.index_1
    df_month = pd.DataFrame(columns=['name'])
    df_month.name = adata1.var.index.values
    df_month.index = df_month.name.values
    df_month.drop(['name'],axis=1,inplace=True)
    var_1 = pd.concat([index_list,df_month],join='outer',axis=1).copy()
    var_1 = var_1.fillna('*')
    var_1['name']=var_1.index
    # 替换index
    var_1.loc[var_1['index_1']!='*','name'] = var_1[var_1['index_1']!='*'].index_new.values
    var_1.index = var_1.name
    adata1.var =var_1
    # 输出结果
    adata1.var_name = var_1.index
    
     # 同上
    index_list = atac2_new.var
    index_list.index = index_list.index_2
    df_month = pd.DataFrame(columns=['name'])
    df_month.name = adata2.var.index.values
    df_month.index = df_month.name.values
    df_month.drop(['name'],axis=1,inplace=True)
    var_2 = pd.concat([index_list,df_month],join='outer',axis=1).copy()
    var_2 = var_2.fillna('*')
    var_2['name']=var_2.index
    var_2.loc[var_2['index_1']!='*','name'] = var_2[var_2['index_1']!='*'].index_new.values
    var_2.index = var_2.name
    adata2.var =var_2
    adata2.var_name = var_2.index
    adata1.var_names_make_unique()
    adata2.var_names_make_unique()
    adata_pair = ad.concat([adata1,adata2],axis=0,join='outer',fill_value=0)
    adata_pair.obs_names_make_unique()
    adata_pair=adata_pair[:,[i for i in adata_pair.var_names if '-' not in i]]


    gc.collect()
    return adata_pair


import torch
import numpy as np
import pandas as pd
import os
import torch.nn as nn


from .neural_net import get_loss

# def process_files(output_folder):
#     smallest_loss = np.Inf
#     best_model_folder_path = None
#     best_mod=None

#     for folder_name in os.listdir(output_folder):
#         folder_path = os.path.join(output_folder, folder_name)

#         if os.path.isdir(folder_path):
#             model_path = os.path.join(folder_path, 'final_model.pt')

#             if os.path.exists(model_path):
#                 try:
#                     mod =  torch.load(model_path)
#                     St=torch.load(os.path.join(folder_path, 'Storch.pt'))
#                     At=torch.load(os.path.join(folder_path, 'Atorch.pt'))
#                     loss = get_loss(mod,St,At)

#                     if loss < smallest_loss:
#                         smallest_loss = loss
#                         best_model_folder_path = folder_path
#                         best_mod=mod
#                 except Exception as e:
#                     raise Exception(f"Error loading model from {model_path}: {str(e)}")
#     print(f'best model: {best_model_folder_path}')
#     if best_model_folder_path:
#         # folder_name = os.path.basename(os.path.dirname(best_model_path))
#         storch_path = os.path.join(best_model_folder_path, 'Storch.pt')
#         atorch_path = os.path.join(best_model_folder_path, 'Atorch.pt')

#         if os.path.exists(storch_path) and os.path.exists(atorch_path):
#             A_torch = torch.load(atorch_path)
#             S_torch = torch.load(storch_path)
            
#             A = A_torch.detach().numpy()
#             S = S_torch.detach().numpy()

#     else:
#         raise Exception("No 'final_model.pt' found in any folder.")

#     return best_mod, A, S

def process_files(output_folder, output_torch=False, epoch_number='final', seed_list=None):
    smallest_loss = np.inf
    best_model_folder_path = None
    best_mod=None

    # Check PyTorch version
    if torch.__version__>='2.6.0':
        is_torch_26_or_later = True
    else:
        is_torch_26_or_later = False

    # only look at specific seeds
    if seed_list is None:
        folder_list=os.listdir(output_folder)
    else:
        folder_list=[f'seed{i}' for i in seed_list]

    for folder_name in folder_list:
        folder_path = os.path.join(output_folder, folder_name)
        if os.path.isdir(folder_path) and 'Storch.pt' in os.listdir(folder_path) and 'Atorch.pt' in os.listdir(folder_path):
            St=torch.load(os.path.join(folder_path, 'Storch.pt'))
            At=torch.load(os.path.join(folder_path, 'Atorch.pt'))

            # check if final_model exists
            if epoch_number!='final':
                final_model_name=f'model_epoch_{epoch_number}.pt'
            else:
                final_model_name='final_model.pt'
                
            final_model_path=os.path.join(folder_path, final_model_name)
            if os.path.exists(final_model_path):
                model_path=final_model_path
            else:
                # find highest epoch model and load
                highest_epoch=-np.inf
                for filename in os.listdir(folder_path):
                    if "model_epoch_" in filename:
                        # Extract the epoch number from the filename
                        epoch_num = int(filename.split('_')[-1][:-3])
                        # Update the highest_epoch and highest_epoch_file if this file has a higher epoch
                        if epoch_num > highest_epoch:
                            highest_epoch = epoch_num
                            highest_epoch_file = filename
                model_path=os.path.join(folder_path, highest_epoch_file)
        
            # Load model based on PyTorch version
            if is_torch_26_or_later:
                try:
                    # First try with weights_only=False
                    mod = torch.load(model_path, weights_only=False)
                except Exception as e:
                    # If that fails, try with safe_globals
                    from torch.serialization import safe_globals
                    from ..gaston.neural_net import GASTON
                    with safe_globals([GASTON]):
                        mod = torch.load(model_path)
            else:
                mod = torch.load(model_path)

            loss = get_loss(mod,St,At)

            # compare against other models with different seeds
            if loss < smallest_loss:
                smallest_loss = loss
                best_model_folder_path = folder_path
                best_mod=mod
                
    print(f'\nbest model: {best_model_folder_path}')
    if best_model_folder_path:
        storch_path = os.path.join(best_model_folder_path, 'Storch.pt')
        atorch_path = os.path.join(best_model_folder_path, 'Atorch.pt')

        if os.path.exists(storch_path) and os.path.exists(atorch_path):
            A_torch = torch.load(atorch_path)
            S_torch = torch.load(storch_path)

            A = A_torch.detach().numpy()
            S = S_torch.detach().numpy()

    else:
        raise Exception("No model found in any folder.")

    if output_torch:
        return best_mod, A, S, A_torch, S_torch
    else:
        return best_mod, A, S

def create_cell_type_df(ct_labels):
    ct_list=np.unique(ct_labels)
    ct_arr=np.zeros( (len(ct_labels), len(ct_list) ))
    
    for i,ct in enumerate(ct_labels):
        ct_arr[i, ct_list==ct]=1
    cell_type_df=pd.DataFrame(ct_arr,columns=ct_list)
    return cell_type_df



"""Spatial analysis utilities."""

from ._cluster import pySTAGATE,clusters,merge_cluster
from ._tangram import Tangram
from ._integrate import pySTAligner,Cal_Spatial_Net
from ._spaceflow import pySpaceFlow
from ._spatrio import CellMap,CellLoc
from ._stt import STT
from ._svg import svg
from ._cast import CAST
from ._tools import *
from ._gaston import GASTON


__all__ = []



"""Bulk omics analysis utilities."""

# Heavy functionality lives in submodules and is imported lazily.
from ._Gene_module import pyWGCNA,readWGCNA
from ._Enrichment import pyGSEA,pyGSE,geneset_enrichment,geneset_plot,geneset_enrichment_GSEA,geneset_plot_multi,enrichment_multi_concat
from ._network import pyPPI,string_interaction,string_map,generate_G
from ._chm13 import get_chm13_gene,find_chm13_gene
from ._Deseq2 import pyDEG,deseq2_normalize,estimateSizeFactors,estimateDispersions,Matrix_ID_mapping,data_drop_duplicates_index
from ._tcga import pyTCGA
from ._combat import batch_correction

__all__ = [
    'pyWGCNA',
    'readWGCNA',
    'pyGSEA',
    'pyGSE',
    'geneset_enrichment',
    'geneset_plot',
    'geneset_enrichment_GSEA',
    'geneset_plot_multi',
    'enrichment_multi_concat',
    'pyPPI',
    'string_interaction',
    'string_map',
    'generate_G',
    'get_chm13_gene',
    'find_chm13_gene',
    'pyDEG',
    'deseq2_normalize',
    'estimateSizeFactors',
    'estimateDispersions',
    'Matrix_ID_mapping',
    'data_drop_duplicates_index',
    'pyTCGA',
    'batch_correction',
    'pyDEG',
    'deseq2_normalize',
    'estimateSizeFactors',
    'estimateDispersions',
    'Matrix_ID_mapping',
    'data_drop_duplicates_index',
    'pyTCGA',
    'batch_correction',
]
